import math as mat
try:
    import pymc3 as pm
    import theano.tensor as t
except ImportError:  # pragma: sin cobertura
    pm = t = None

# Funciones auxiliares para ecuaciones.

# Un diccionario con conversiones de funciones reconocidas. Si quieres activar más funciones, agregarlas aqui.
_dic_funs = {
    'mín': {'vensim': 'MIN', 'pm': pm.math.minimum if pm is not None else None, 'python': min},
    'máx': {'vensim': 'MAX', 'pm': pm.math.maximum if pm is not None else None, 'python': max},
    'abs': {'vensim': 'ABS', 'pm': pm.math.abs_ if pm is not None else None, 'python': abs},
    'exp': {'vensim': 'EXP', 'pm': pm.math.exp if pm is not None else None, 'python': mat.exp},
    'ent': {'vensim': 'INTEGER', 'pm': pm.math.floor if pm is not None else None, 'python': int},
    'rcd': {'vensim': 'SQRT', 'pm': pm.math.sqrt if pm is not None else None, 'python': mat.sqrt},
    'ln': {'vensim': 'LN', 'pm': pm.math.log if pm is not None else None, 'python': mat.log},
    'log': {'vensim': 'LOG', 'pm': None if pm is None else lambda x: pm.math.log(x) / mat.log(10), 'python': mat.log10},
    'sin': {'vensim': 'SIN', 'pm': pm.math.sin if pm is not None else None, 'python': mat.sin},
    'cos': {'vensim': 'COS', 'pm': pm.math.cos if pm is not None else None, 'python': mat.cos},
    'tan': {'vensim': 'TAN', 'pm': pm.math.tan if pm is not None else None, 'python': mat.tan},
    'sinh': {'vensim': 'SINH', 'pm': pm.math.sinh if pm is not None else None, 'python': mat.sinh},
    'cosh': {'vensim': 'COSH', 'pm': pm.math.cosh if pm is not None else None, 'python': mat.cosh},
    'tanh': {'vensim': 'TANH', 'pm': pm.math.tanh if pm is not None else None, 'python': mat.tanh},
    'asin': {'vensim': 'ARCSIN', 'pm': t.arcsin if t is not None else None, 'python': mat.asin},
    'acos': {'vensim': 'ARCCOS', 'pm': t.arccos if t is not None else None, 'python': mat.acos},
    'atan': {'vensim': 'ARCTAN', 'pm': t.arctan if t is not None else None, 'python': mat.atan},
    'si_sino': {'vensim': 'IF THEN ELSE', 'pm': pm.math.switch if pm is not None else None,
                'python': lambda cond, si, sino: si if cond else sino}
}

# Diccionario de operadores. Notar que algunos se traducen por funciones en PyMC3.
_dic_ops = {
    '>': {'vensim': '>', 'pm': pm.math.gt if pm is not None else None},
    '<': {'vensim': '<', 'pm': pm.math.lt if pm is not None else None},
    '>=': {'vensim': '>=', 'pm': pm.math.ge if pm is not None else None},
    '<=': {'vensim': '<=', 'pm': pm.math.le if pm is not None else None},
    '==': {'vensim': '=', 'pm': pm.math.eq if pm is not None else None},
    '!=': {'vensim': '<>', 'pm': pm.math.neq if pm is not None else None},
    '+': {'vensim': '+'},
    '-': {'vensim': '-'},
    '*': {'vensim': '*'},
    '^': {'vensim': '^'},
    '/': {'vensim': '/'}
}


def conv_fun(fun, dialecto_orig, dialecto_final):
    """
    Traduce una función a otro dialecto.

    Parameters
    ----------
    fun: str
        La función para traducir.
    dialecto_orig: str
        El dialecto original de la función.
    dialecto_final: str
        El dialecto final deseado.

    Returns
    -------
    str | Callable
        La función traducida.
    """

    if dialecto_final == dialecto_orig:
        return fun
    if dialecto_orig == 'tinamït':
        return _dic_funs[fun][dialecto_final]
    else:
        if dialecto_final == 'tinamït':
            return next(ll for ll, d in _dic_funs.items() if d[dialecto_orig] == fun)
        else:
            return next(d[dialecto_final] for ll, d in _dic_funs.items() if d[dialecto_orig] == fun)


def conv_op(oper, dialecto_orig, dialecto_final):
    """
    Traduce un operador a otro dialecto.

    Parameters
    ----------
    oper: str
        El operador para traducir.
    dialecto_orig: str
        El dialecto original del operador.
    dialecto_final: str
        El dialecto final deseado.

    Returns
    -------
    str | Callable
        El operador traducido.
    """

    if dialecto_final == dialecto_orig:
        return oper
    if dialecto_orig == 'tinamït':
        return _dic_ops[oper][dialecto_final]
    else:
        if dialecto_final == 'tinamït':
            return next(ll for ll, d in _dic_ops.items() if d[dialecto_orig] == oper)
        else:
            return next(d[dialecto_final] for ll, d in _dic_ops.items() if d[dialecto_orig] == oper)



from tinamit.envolt.bf import ModeloBF
from tinamit.mod import VariablesMod, Variable


class PruebaBF(ModeloBF):

    def __init__(símismo, unid_tiempo='mes', nombre='bf'):
        símismo.unid_tiempo = unid_tiempo
        super().__init__(nombre=nombre, variables=símismo._gen_vars(unid_tiempo))

    @staticmethod
    def _gen_vars(unid_t):
        return VariablesMod([
            Variable('Lluvia', unid='m3/'+unid_t, ingr=False, egr=True, inic=1, líms=(0, None)),
            Variable('Lago', unid='m3', ingr=True, egr=False, inic=1000, líms=(0, None)),
        ])

    def incrementar(símismo, rebanada):
        lago = símismo.variables['Lago'].obt_val()
        símismo.variables['Lluvia'].poner_val(lago / 120 * rebanada.n_pasos)
        super().incrementar(rebanada)

    def unidad_tiempo(símismo):
        return símismo.unid_tiempo

    def paralelizable(símismo):
        return True



import os
import unittest

import numpy as np
import numpy.testing as npt

from tinamit.cositas import guardar_json, jsonificar, cargar_json


def verificar_leer_ingr(caso, cls):
    """
    Verifica que una envoltura lee bien sus datos de ingreso.

    Parameters
    ----------
    caso: unittest.TestCase
        El caso de prueba.
    cls:
        La clase del modelo para comprobar.

    """
    info_prb = cls.prb_ingreso()
    if info_prb:
        arch, f_leer = info_prb
        caso.maxDiff = None
        dic = {str(vr): jsonificar(vr.__dict__) for vr in f_leer(arch)}

        dir_, nombre = os.path.split(arch)
        ref = {str(ll): jsonificar(v) for ll, v in _obt_ref(os.path.join(dir_, 'ref', nombre + '.json'), dic).items()}

        caso.assertSetEqual(set(dic), set(ref))
        for vr in ref:
            caso.assertEqual(dic[vr]['unid'], ref[vr]['unid'], vr)
            caso.assertTupleEqual(tuple(dic[vr]['líms']), tuple(ref[vr]['líms']), vr)
            npt.assert_equal(dic[vr]['inic'], ref[vr]['inic'], err_msg=vr)


def verificar_leer_egr(caso, cls):
    """
    Verifica que una envoltura lee bien sus datos de egreso.

    Parameters
    ----------
    caso: unittest.TestCase
        El caso de prueba.
    cls:
        La clase del modelo para comprobar.

    """
    info_prb = cls.prb_egreso()
    if info_prb:
        arch, f_leer = info_prb
        dic = {vr: np.array(vl) for vr, vl in f_leer(arch).items()}

        dir_, nombre = os.path.split(arch)
        ref = _obt_ref(os.path.join(dir_, 'ref', nombre + '.json'), dic)
        caso.assertSetEqual(set(dic), set(ref))
        for vr in ref:
            npt.assert_equal(dic[vr], ref[vr], err_msg=vr)


def verificar_simul(caso, cls):
    """
    Verifica que una envoltura simula bien. No correrá si no está instalada la envoltura.

    Parameters
    ----------
    caso: unittest.TestCase
        El caso de prueba.
    cls:
        La clase del modelo para comprobar.

    """
    arch_ingr = cls.prb_simul()
    if arch_ingr and cls.instalado():
        mod = cls(arch_ingr)
        res = mod.simular(2).a_dic()

        dir_, nombre = os.path.split(arch_ingr)

        ref = _obt_ref(os.path.join(dir_, 'ref', nombre + '.simul.json'), res)
        caso.assertDictEqual(res, ref)


def _obt_ref(arch, d_auto):
    if not os.path.isfile(arch):
        guardar_json(jsonificar(d_auto), arch)

    return cargar_json(arch)



from ...mod import Modelo


class ModeloDS(Modelo):
    ext = []

    def __init__(símismo, variables, nombre='mds'):
        super().__init__(variables, nombre)

    def unidad_tiempo(símismo):
        raise NotImplementedError



import json
import os
import shutil
import tempfile
from datetime import datetime, date
from warnings import warn as avisar

import numpy as np
from chardet import UniversalDetector


def detectar_codif(archivo, máx_líneas=None, cortar=None):
    """
    Detecta la codificación de un fuente. (Necesario porque todavía existen programas dinosaurios que no entienden
    los milagros de unicódigo.)

    Parameters
    ----------
    archivo : str
        La dirección del fuente.
    máx_líneas : int
        El número máximo de líneas para pasar al detector. Por ejemplo, si únicamente la primera línea de tu documento
        tiene palabras (por ejemplo, un fuente .csv), no hay razón de seguir analizando las otras líneas.
    cortar : str
        Hasta dónde hay que leer cada línea. Es útil si tienes un csv donde únicamente la primera columna
        contiene texto.

    Returns
    -------
    str
        La codificación más probable.

    """

    detector = UniversalDetector()
    with open(archivo, 'rb') as d:
        for í, línea in enumerate(d.readlines()):

            if cortar is None or cortar.encode() not in línea:
                detector.feed(línea)  # Pasar la próxima línea al detector
            else:
                detector.feed(línea.split(cortar.encode())[0])

            # Parar si alcanzamos el máximo de líneas
            if máx_líneas is not None and í >= (máx_líneas - 1):
                break

            if detector.done:
                break  # Para si el detector ya está seguro

    detector.close()  # Cerrar el detector

    return detector.result['encoding']  # Devolver el resultado


def valid_nombre_arch(nombre):
    """
    Valida el nombre de un fuente.

    Parameters
    ----------
    nombre : str
        El nombre propuesto para el fuente.
    Returns
    -------
    str
        Un nombre de fuente válido.
    """

    for x in ['\\', '/', '|', ':' '*', '?', '"', '>', '<']:
        nombre = nombre.replace(x, '_')

    return nombre.strip()


def arch_más_recién(arch1, arch2):
    return os.path.getmtime(arch1) > os.path.getmtime(arch2)


def guardar_json(obj, arch):
    """
    Guarda un fuente json, sin riesgo de corrumpir el fuente si se interrumpe el programa mientras escribía al disco.

    Parameters
    ----------
    obj : list | dict
        Objeto compatible json.
    arch : str
        El fuente en el cual hay que guardar el objeto json.

    """
    nmbr, ext = os.path.splitext(arch)
    if not len(ext):
        arch = nmbr + '.json'

    contenido = json.dumps(obj, ensure_ascii=False, sort_keys=True, indent=2)
    guardar_archivo(arch, contenido=contenido)


def guardar_archivo(arch, contenido):
    """
    Guarda un documento de manera segura, sin riesgo de corrumpir el archivo existente si se interrumpe el proceso.

    Parameters
    ----------
    arch : str
        El archivo que hay que escribir.
    contenido : str
        El contenido deseado del archivo.

    """

    with tempfile.NamedTemporaryFile('w', encoding='UTF-8', delete=False) as temp:
        # Escribimos primero a un fuente temporario para evitar de corrumpir nuestro fuente principal si el programa
        # se interrumpe durante la operación.
        temp.write(contenido)

        # Crear el directorio, si necesario
        direc = os.path.split(arch)[0]
        if len(direc) and not os.path.isdir(direc):  # pragma: sin cobertura
            os.makedirs(os.path.split(arch)[0])

    # Después de haber escrito el fuente, ya podemos cambiar el nombre sin riesgo.
    try:

        if os.path.splitdrive(temp.name)[0] == os.path.splitdrive(arch)[0]:
            os.replace(temp.name, arch)
        else:
            # En casos de documentos en distintos discos, debemos emplear `shutil` en vez.
            shutil.move(temp.name, arch)

    except (PermissionError, FileNotFoundError, OSError):  # pragma: sin cobertura
        # Necesario en el caso de corridas en paralelo en Windows. Sin este, la reimportación de Tinamït ocasionada
        # por varias corridas paralelas al mismo tiempo puede causar que el mismo documento se escriba por dos procesos
        # al mismo tiempo, el cual trava el sistema.
        avisar(arch)


def cargar_json(arch, codif='UTF-8'):
    """
    Cargar un fuente json.

    Parameters
    ----------
    arch : str
        El fuente en el cual se encuentra el objeto json.
    codif : str
        La codificación del fuente.

    Returns
    -------
    dict | list
        El objeto json.

    """

    nmbr, ext = os.path.splitext(arch)
    if not len(ext):
        arch = nmbr + '.json'

    with open(arch, 'r', encoding=codif) as d:
        return json.load(d)


def jsonificar(dic, redond=None):
    nuevo = {}
    for ll, v in dic.items():
        if isinstance(v, dict):
            nuevo[ll] = jsonificar(v, redond=redond)
        elif isinstance(v, np.ndarray):
            if redond:
                v = v.round(redond)
            nuevo[ll] = v.tolist()
        else:
            nuevo[ll] = v

    return nuevo


def numpyficar(dic):
    nuevo = {}
    for ll, v in dic.items():
        if isinstance(v, dict):
            nuevo[ll] = numpyficar(v)
        elif isinstance(v, list):
            nuevo[ll] = np.array(v)
        else:
            nuevo[ll] = v
    return nuevo


def _gen_fecha(f):
    if f is None:
        return
    elif isinstance(f, str):
        return datetime.strptime(f, '%Y-%m-%d')
    elif isinstance(f, datetime):
        return f
    elif isinstance(f, date):
        return datetime(f.year, f.month, f.day)
    else:
        raise TypeError(type(f))



"""
Python model "prueba_senc_mdl.py"
Translated using PySD version 0.10.0
"""
from __future__ import division
import numpy as np
from pysd import utils
import xarray as xr

from pysd.py_backend.functions import cache
from pysd.py_backend import functions

_subscript_dict = {}

_namespace = {
    'TIME': 'time',
    'Time': 'time',
    'Aleatorio': 'aleatorio',
    'Evaporación': 'evaporación',
    'Flujo río': 'flujo_río',
    'Lago': 'lago',
    'Lluvia': 'lluvia',
    'Nivel lago inicial': 'nivel_lago_inicial',
    'FINAL TIME': 'final_time',
    'INITIAL TIME': 'initial_time',
    'SAVEPER': 'saveper',
    'TIME STEP': 'time_step'
}

__pysd_version__ = "0.10.0"

__data = {'scope': None, 'time': lambda: 0}


def _init_outer_references(data):
    for key in data:
        __data[key] = data[key]


def time():
    return __data['time']()


@cache('step')
def aleatorio():
    """
    Real Name: b'Aleatorio'
    Original Eqn: b'GAME ( 0)'
    Units: b'Sdmn'
    Limits: (0.0, 1.0)
    Type: component

    b'Un variable aleatorio que debe provenir del modelo BP.'
    """
    return (0)


@cache('step')
def evaporación():
    """
    Real Name: b'Evaporaci\\xf3n'
    Original Eqn: b'0.1*Lago'
    Units: b'm3/mes'
    Limits: (0.0, None)
    Type: component

    b'La evaporaci\\xf3n del lago, fijo al 10%.'
    """
    return 0.1 * lago()


@cache('step')
def flujo_río():
    """
    Real Name: b'Flujo r\\xedo'
    Original Eqn: b'Lluvia'
    Units: b'm3/mes'
    Limits: (0.0, None)
    Type: component

    b'El flujo de agua en el r\\xedo.'
    """
    return lluvia()


@cache('step')
def lago():
    """
    Real Name: b'Lago'
    Original Eqn: b'INTEG ( Flujo r\\xedo-Evaporaci\\xf3n, Nivel lago inicial)'
    Units: b'm3'
    Limits: (0.0, None)
    Type: component

    b'La cantidad de agua en el lago.'
    """
    return _integ_lago()


@cache('step')
def lluvia():
    """
    Real Name: b'Lluvia'
    Original Eqn: b'GAME ( 10)'
    Units: b'm3/mes'
    Limits: (0.0, None)
    Type: component

    b'La cantidad de lluvia que cae por mes.'
    """
    return (10)


@cache('run')
def nivel_lago_inicial():
    """
    Real Name: b'Nivel lago inicial'
    Original Eqn: b'1500'
    Units: b'm3'
    Limits: (0.0, None)
    Type: constant

    b'La cantidad inicial  de ag   ua en el lago.'
    """
    return 1500


@cache('run')
def final_time():
    """
    Real Name: b'FINAL TIME'
    Original Eqn: b'200'
    Units: b'mes'
    Limits: (None, None)
    Type: constant

    b'The final time for the simulation.'
    """
    return 200


@cache('run')
def initial_time():
    """
    Real Name: b'INITIAL TIME'
    Original Eqn: b'0'
    Units: b'mes'
    Limits: (None, None)
    Type: constant

    b'The initial time for the simulation.'
    """
    return 0


@cache('step')
def saveper():
    """
    Real Name: b'SAVEPER'
    Original Eqn: b'TIME STEP'
    Units: b'mes'
    Limits: (0.0, None)
    Type: component

    b'The frequency with which output is stored.'
    """
    return time_step()


@cache('run')
def time_step():
    """
    Real Name: b'TIME STEP'
    Original Eqn: b'1'
    Units: b'mes'
    Limits: (0.0, None)
    Type: constant

    b'The time step for the simulation.'
    """
    return 1


_integ_lago = functions.Integ(lambda: flujo_río() - evaporación(), lambda: nivel_lago_inicial())



from ast import literal_eval

import numpy as np
import pandas as pd

from tinamit.calibs.sintx.ec import Ecuación
from tinamit.envolt.mds.pysd._funcs import decodar
from ._funcs import gen_mod_pysd, obt_paso_mod_pysd
from ._vars import VarPySDAuxiliar, VarPySDNivel, VarPySDConstante, VariablesPySD
from .._envolt import ModeloDS


class ModeloPySD(ModeloDS):
    """
    Envoltura para modelos PySD.
    """
    ext = ['.mdl', '.xmile', '.xml', '.py']

    def __init__(símismo, archivo, nombre='mds'):
        símismo.mod = gen_mod_pysd(archivo)
        símismo.vars_para_cambiar = {}
        símismo.archivo = archivo

        símismo.cont_simul = False
        símismo.paso_act = 0

        super().__init__(variables=_gen_vars(símismo.mod), nombre=nombre)

    def iniciar_modelo(símismo, corrida):
        # Poner los variables y el tiempo a sus valores iniciales
        símismo.mod.reload()
        símismo.mod.initialize()
        símismo.paso_act = 0

        símismo.cont_simul = False

        super().iniciar_modelo(corrida)

    def incrementar(símismo, rebanada):

        símismo.paso_act += rebanada.n_pasos
        devolv = [símismo.paso_act - 1, símismo.paso_act]
        res = símismo.mod.run(
            initial_condition='current' if símismo.cont_simul else 'original', params=símismo.vars_para_cambiar,
            return_timestamps=devolv, return_columns=[v.nombre_py for v in rebanada.resultados.variables()]
        )

        # Guardar valores iniciales
        for v in rebanada.resultados.variables():
            rebanada.resultados[v].vals[símismo.paso_act - 1] = res[v.nombre_py][símismo.paso_act - 1]

        símismo.cont_simul = True
        símismo.variables.cambiar_vals({v: res[v.nombre_py].values[-1] for v in rebanada.resultados.variables()})

        símismo.vars_para_cambiar.clear()
        super().incrementar(rebanada)

    def _correr_hasta_final(símismo):
        t = símismo.corrida.t

        t_inic_mod = símismo.mod.time._t
        eje_pysd = np.arange(
            t_inic_mod, t_inic_mod + t.n_pasos * t.tmñ_paso + 1, t.guardar_cada * t.tmñ_paso
        )

        if símismo.corrida.extern is not None:
            vals_extern = símismo.corrida.extern.obt_vals(t.eje(), var=símismo.variables)
            paráms = {
                vr: vl.squeeze().to_pandas()
                for vr, vl in vals_extern.items()
            }
            for ll, v in paráms.items():
                if isinstance(v, pd.Series):
                    v.index = eje_pysd[t.eje().isin(v.index.values)]
        else:
            paráms = {}

        if símismo.corrida.clima is not None:
            vars_clima = símismo.corrida.clima.obt_todos_vals(t.eje(), vars_clima=símismo.vars_clima)
            paráms.update(vars_clima)

        símismo._proc_paráms(paráms)

        res_pysd = símismo.mod.run(
            params=paráms,
            return_timestamps=eje_pysd
        )
        return {vr: res_pysd[str(vr)].values for vr in símismo.corrida.resultados.variables()}

    def unidad_tiempo(símismo):
        return obt_paso_mod_pysd(símismo.mod)

    def cambiar_vals(símismo, valores):
        símismo.vars_para_cambiar.update({vr: float(vl) for vr, vl in valores.items()})
        super().cambiar_vals(valores)

    def paralelizable(símismo):
        return True

    def _proc_paráms(símismo, prms):
        for p in list(prms):
            v = símismo.variables[p]
            if isinstance(v, VarPySDNivel):
                val = prms.pop(p)
                if v.var_inic:
                    prms[v.var_inic.nombre] = val


def _gen_vars(mod):
    l_vars = []

    internos = ['FINAL TIME', 'TIME STEP', 'SAVEPER', 'INITIAL TIME']
    for i, f in mod.doc().iterrows():

        nombre = f['Real Name']
        if f['Type'] == 'lookup' or nombre in internos:
            continue

        nombre_py = f['Py Name']
        unid = decodar(f['Unit'])
        líms = literal_eval(f['Lims'])
        ec = decodar(f['Eqn'])
        info = decodar(f['Comment'])
        obj_ec = Ecuación(ec, dialecto='vensim')
        parientes = {v for v in obj_ec.variables() if v not in internos}
        inic = getattr(mod.components, nombre_py)()

        try:
            getattr(mod.components, '_integ_' + nombre_py)
            nivel = True

        except AttributeError:
            nivel = False
        if nivel:
            var = VarPySDNivel(
                nombre, nombre_py=nombre_py, unid=unid, ec=ec, parientes=parientes, inic=inic, líms=líms, info=info
            )
        elif len(parientes):
            var = VarPySDAuxiliar(
                nombre, nombre_py=nombre_py, unid=unid, ec=ec, parientes=parientes, inic=inic, líms=líms, info=info
            )
        else:
            var = VarPySDConstante(
                nombre, nombre_py=nombre_py, unid=unid, ec=ec, parientes=parientes, inic=inic, líms=líms, info=info
            )

        l_vars.append(var)

    for vr in l_vars:
        if isinstance(vr, VarPySDNivel):
            # para hacer: no funciona con XMILE, solamente con vensim (.mdl)
            args_integ = Ecuación(vr.ec).sacar_args_func('INTEG', 2)
            if args_integ:
                arg_inic = args_integ[1]
                vr.estab_var_inic(next((
                    v for v in l_vars if v.nombre == arg_inic
                ), None))

    return VariablesPySD(l_vars)



import itertools

from tinamit.config import _


class Corrida(object):
    def __init__(símismo, nombre, t, extern, vars_mod, vars_interés, clima):
        símismo.nombre = nombre
        símismo.t = t
        símismo.extern = extern

        símismo.variables = vars_mod

        símismo.clima = clima

        símismo.resultados = vars_mod.gen_res(
            nombre=nombre, t=t, vars_interés=vars_interés
        )

    def actualizar_res(símismo):
        if símismo.t.t_guardar():
            símismo.resultados.actualizar()

    def obt_extern_act(símismo, var=None):
        if símismo.extern:
            return {vr: vl.values[0] for vr, vl in símismo.extern.obt_vals(símismo.t, var=var).items()}

    def __str__(símismo):
        return símismo.nombre


class Rebanada(object):
    def __init__(símismo, n_pasos, resultados):
        símismo.n_pasos = n_pasos

        símismo.resultados = resultados


class PlantillaOpsSimulGrupo(object):
    """
    Clase pariente para especificaciones de corridas en grupo.
    """
    def __init__(símismo, **argsll):
        símismo.opciones = argsll

        for ll, v in símismo.opciones.items():
            if not isinstance(v, list):
                símismo.opciones[ll] = [v]

    def __iter__(símismo):
        raise NotImplementedError


class OpsSimulGrupoCombin(PlantillaOpsSimulGrupo):
    """
    Corridas en grupo que generan todas las combinaciones posibles de las opciones de simulación especificadas.
    """
    
    def __init__(símismo, t, extern=None, clima=None, vars_interés=None, nombre='Tinamït'):
        símismo.nombre = nombre
        símismo.vars_interés = vars_interés
        super().__init__(t=t, extern=extern, clima=clima)

    def __iter__(símismo):
        ops = itertools.product(*símismo.opciones.values())
        nmbs = ['t', 'extern', 'clima']
        for i, op in enumerate(ops):
            yield {
                **{n: o for n, o in zip(nmbs, op)},
                'nombre': '{}_{}'.format(símismo.nombre, str(i)),
                'vars_interés': símismo.vars_interés
            }


class OpsSimulGrupo(PlantillaOpsSimulGrupo):
    """
    Corridas en grupo sin combinaciones (las primeras opciones de cada parámetro se corrirán juntas en una
    simulación, seguidas por las segundas, etc.).
    """

    def __init__(símismo, t, extern=None, clima=None, vars_interés=None, nombre='Tinamït'):
        símismo.nombre = nombre
        símismo.vars_interés = vars_interés

        args = {
            't': t, 'extern': extern, 'clima': clima
        }
        if not isinstance(nombre, str):
            args['nombre'] = nombre

        super().__init__(**args)

        tmñs = {len(o) for o in símismo.opciones.values() if len(o) != 1}

        if len(tmñs) > 1:
            raise ValueError(_('Todas las opciones deben tener el mismo número de valores.'))
        elif not tmñs:
            símismo.tmñ_ops = 1
        else:
            símismo.tmñ_ops = list(tmñs)[0]

        if 'nombre' not in símismo.opciones:
            símismo.opciones['nombre'] = ['{}_{}'.format(nombre, str(i)) for i in range(símismo.tmñ_ops)]

    def __iter__(símismo):
        for i in range(símismo.tmñ_ops):
            yield {
                **{n: op[i] if len(op) > 1 else op[0] for n, op in símismo.opciones.items()},
                'vars_interés': símismo.vars_interés
            }



from tinamit.config import _
from ._res import ResultadosConectado
from ..mod import VariablesMod


class VariablesConectado(VariablesMod):

    def __init__(símismo, modelos):
        super().__init__(variables=[])
        símismo.modelos = modelos
        símismo.variables = {str(m): m.variables for m in modelos}

    def gen_res(símismo, nombre, t, vars_interés):
        return ResultadosConectado(
            nombre=nombre, t=t, modelos=símismo.modelos, vars_interés=símismo._resolver_vars(vars_interés)
        )

    def _resolver(símismo, var):
        try:
            return next((m, v_m[var]) for m, v_m in símismo.variables.items() if var in v_m)
        except StopIteration:
            raise KeyError(_('Variable {var} no existe.').format(var))

    def __getitem__(símismo, itema):
        mod, var = símismo._resolver(itema)
        return símismo.variables[mod][var]

    def __setitem__(símismo, llave, valor):
        mod, var = símismo._resolver(llave)
        símismo.variables[mod][var] = valor

    def __iter__(símismo):
        for m, v_m in símismo.variables.items():
            for v in v_m:
                yield v

    def __contains__(símismo, itema):
        try:
            símismo._resolver(itema)
            return True
        except KeyError:
            return False



import re
import tempfile

import numpy as np
import pandas as pd
import spotpy as sp
import xarray as xr
import platform

from tinamit.cositas import detectar_codif
from tinamit.datos.bd import BD

from ._utils import calc_máx_trz, algs_spotpy, eval_funcs


class CalibradorMod(object):
    """
    Clase pariente para cada calibrador de modelo.
    """

    def __init__(símismo, mod):
        símismo.mod = mod

    def calibrar(símismo, líms_paráms, datos, método='epm', n_iter=300, vars_obs=None):
        """
        Efectuar la calibración.

        Parameters
        ----------
        líms_paráms: dict
            Diccionario de cada parámetro con sus límites teoréticos.
        datos: xr.Dataset or xr.DataArray or str or pd.DataFrame or dict or Fuente or list or BD
            Los datos para la calibración.
        método
        n_iter
        vars_obs

        Returns
        -------

        """
        # Para hacer: limpiar comunicación de datos entre calibrador y CalibradorGeog, y también con validadores
        if isinstance(datos, xr.Dataset):
            obs = datos
        else:
            datos = datos if isinstance(datos, BD) else BD(datos)
            obs = datos.obt_vals(vars_obs)

        return símismo._efec_calib(líms_paráms=líms_paráms, método=método, n_iter=n_iter, obs=obs)

    def _efec_calib(símismo, líms_paráms, método, n_iter, obs):
        """
        Efectua la calibración.

        Parameters
        ----------
        líms_paráms: dict
            Diccionario de cada parámetro con sus límites teoréticos.
        método
        n_iter
        obs: xr.Dataset

        Returns
        -------

        """
        raise NotImplementedError


class CalibradorModSpotPy(CalibradorMod):

    def _efec_calib(símismo, líms_paráms, método, n_iter, obs):

        temp = tempfile.NamedTemporaryFile('w', encoding='UTF-8', prefix='CalibTinamït_')

        mod_spotpy = _ModSpotPy(mod=símismo.mod, líms_paráms=líms_paráms, obs=obs)

        alg = algs_spotpy[método.lower()] if isinstance(método, str) else método
        muestreador = alg(mod_spotpy, dbname=temp.name, dbformat='csv')
        if método is sp.algorithms.dream:
            muestreador.sample(repetitions=n_iter, runs_after_convergence=200)
        else:
            muestreador.sample(n_iter)

        codif = detectar_codif(temp.name + '.csv', máx_líneas=1) if platform.system() == 'Windows' else 'utf8'
        egr_spotpy = pd.read_csv(temp.name + '.csv', encoding=codif)

        cols_prm = [c for c in egr_spotpy.columns if c.startswith('par')]
        probs = egr_spotpy['like1']

        if método in (sp.algorithms.dream, sp.algorithms.mcmc):
            trzs = {p: egr_spotpy[p][-200:] for p in cols_prm}
        else:
            buenas = (probs >= 0.80)
            trzs = {p: egr_spotpy[p][buenas] for p in cols_prm}

        res = {}
        for p in líms_paráms:
            col_p = ('par' + p).replace(' ', '_')
            res[p] = {
                'dist': trzs[col_p], 'cumbre': calc_máx_trz(trzs[col_p]),
                'mejor': egr_spotpy[col_p][np.argmax(probs.values)]
            }

        return res


class _ModSpotPy(object):
    def __init__(símismo, mod, líms_paráms, obs, f_obj='ens'):
        """

        Parameters
        ----------
        mod : Modelo.Modelo
        líms_paráms : dict
        obs: xr.Dataset
        """

        símismo.paráms = [
            sp.parameter.Uniform(re.sub('\W|^(?=\d)', '_', p), low=d[0], high=d[1], optguess=(d[0] + d[1]) / 2)
            for p, d in líms_paráms.items()
        ]
        símismo.nombres_paráms = list(líms_paráms)
        símismo.mod = mod
        símismo.vars_interés = sorted(list(obs.data_vars))
        símismo.t_final = len(obs['n']) - 1  # para hacer: arreglar

        símismo.mu_obs = símismo._aplastar(obs.mean())
        símismo.sg_obs = símismo._aplastar(obs.std())
        símismo.obs_norm = símismo._aplastar((obs - obs.mean()) / obs.std())

        símismo.func = eval_funcs[f_obj.lower()]

    def parameters(símismo):
        return sp.parameter.generate(símismo.paráms)

    def simulation(símismo, x):
        res = símismo.mod.simular(
            t=símismo.t_final, vars_interés=símismo.vars_interés, extern=dict(zip(símismo.nombres_paráms, x))
        )
        m_res = np.array([res[v].vals for v in símismo.vars_interés]).T

        return ((m_res - símismo.mu_obs) / símismo.sg_obs).T.ravel()

    def evaluation(símismo):
        return símismo.obs_norm

    def objectivefunction(símismo, simulation, evaluation, params=None):
        return símismo.func(evaluation, simulation, f=None)

    def _aplastar(símismo, datos):
        if isinstance(datos, xr.Dataset):
            return np.array([datos[v].values.ravel() for v in símismo.vars_interés]).ravel()
        elif isinstance(datos, dict):
            return np.array([datos[v].ravel() for v in sorted(datos)]).ravel()



import numpy as np
import scipy.stats as estad
import spotpy
from spotpy import objectivefunctions as spt_f, likelihoods as spt_l


def calc_máx_trz(trz):
    if len(trz) == 1:
        return trz[0]

    # Normalizar el rango (necesario para las funciones que siguen, si es muy grande)
    sig = np.std(trz)
    mu = np.mean(trz)
    trz_norm = (trz - mu) / sig

    # Intentar calcular la densidad máxima.
    try:
        # Se me olvidó cómo funciona esta parte.
        fdp = estad.gaussian_kde(trz)
        x = np.linspace(trz_norm.min(), trz_norm.max(), 1000)
        máx = x[np.argmax(fdp.evaluate(x))]

        # De-normalizar
        return máx * sig + mu

    except BaseException:
        return np.nan


def _anlz_tendencia(o, s, f):
    # para hacer: con eje tiempo
    x = (f - f[0]).days
    pend_obs = estad.linregress(x, o)[0]
    pend_sim = estad.linregress(x, s)[0]

    return {'obs': pend_obs, 'sim': pend_sim, 'correcta': (pend_obs > 0) is (pend_sim > 0)}


algs_spotpy = {
    'fast': spotpy.algorithms.fast,
    'dream': spotpy.algorithms.dream,
    'cm': spotpy.algorithms.mc,
    'cmmc': spotpy.algorithms.mcmc,
    'epm': spotpy.algorithms.mle,
    'mhl': spotpy.algorithms.lhs,
    'as': spotpy.algorithms.sa,
    'sceua': spotpy.algorithms.sceua,
    'erop': spotpy.algorithms.rope,
    'caa': spotpy.algorithms.abc,
    'fscabc': spotpy.algorithms.fscabc,
    'bdd': spotpy.algorithms.dds
}

eval_funcs = {
    'ens': lambda o, s, f: spt_f.nashsutcliffe(o, s),
    'rcep': lambda o, s, f: -spt_f.rmse(o, s),
    'corresp': lambda o, s, f: spt_f.agreementindex(o, s),
    'ekg': lambda o, s, f: spt_f.kge(o, s),
    'r2': lambda o, s, f: spt_f.rsquared(o, s),
    'rcnep': lambda o, s, f: -spt_f.rrmse(o, s),
    'log p': lambda o, s, f: spt_f.log_p(o, s),
    'verosimil_gaus': lambda o, s, f: spt_l.gaussianLikelihoodMeasErrorOut(o, s),
    'tendencia': _anlz_tendencia
}



import unittest

import tinamit.unids.trads as trads
from tinamit.unids.conv import convertir


class Test_ConvertirUnidades(unittest.TestCase):
    """
    Comprobar la conversión de unidades.
    """

    def test_convertir_unids_equivalentes(símismo):
        # unids equivalentes deberían dar un factor de conversión de 1.
        símismo.assertEqual(convertir('m2/d', 'm3/d/m'), 1)

    def test_convertir_unids_iguales(símismo):
        # unids idénticas deberían dar un factor de conversión de 1.
        símismo.assertEqual(convertir('colibrís', 'colibrís'), 1)

    def test_convertir_con_traducción_necesaria(símismo):
        # Registrar traducciones
        trads.agregar_trad('pound', trad='paj', leng_trad='cak', guardar=False)
        trads.agregar_trad('paj', trad='libra', leng_trad='es', leng_orig='cak', guardar=False)
        de = 'paj'
        a = 'libras'

        # Deberían ser igual ahora.
        símismo.assertEqual(convertir(de, a), 1)

    def test_convertir_con_traducción_desconocida(símismo):
        # Aunque no sepa lo que es una "estación", el programa debería detectar que es lo mismo en ambos y solamente
        # convertir los m3 a cm3.
        símismo.assertAlmostEqual(convertir('m3/estación', 'cm3/estación'), 100 ** 3)

    def test_convertir_con_valor(símismo):
        # Comprobar conversión real.
        símismo.assertEqual(convertir('km', 'm', val=2), 2000)

    def test_conversión_imposible(símismo):
        """
        ¡Comparemos bananas con naranjas!
        """
        símismo.assertRaises(ValueError, lambda: convertir('bananas', 'naranjas'))

        # Y un error de dimensionalidad.
        símismo.assertRaises(ValueError, lambda: convertir('m3', 'm2'))

    def test_convertir_plurial_erróneo(símismo):
        """
        Intentemos convertir un nombre que parece ser forma plurial ("mes" podría reconocerse como el plurial
        de "m"), pero no lo es.
        """
        símismo.assertEqual(convertir('años', 'mes'), 12)


class Test_TradsUnidades(unittest.TestCase):
    """
    Pruebas de funciones de traducción de unidades.
    """

    def test_traducir(símismo):
        """
        Comprobar que podemos traducir.
        """

        res = trads.trad_unid('year', leng_final='es', leng_orig='en')
        símismo.assertEqual(res, 'año')

    def test_traducir_sin_especificar_leng(símismo):
        """
        Comprobar que podamos adividar la lengua original de una unidad.
        """

        res = trads.trad_unid('year', leng_final='es')
        símismo.assertEqual(res, 'año')

    def test_trad_unidad_errónea(símismo):
        """
        Comprobar que devolvamos aviso para unidades que no existen en la lengua especificada.
        """

        res = trads.trad_unid('paj', leng_final='fr', leng_orig='es')

        # Asegurarque que devolvimos el valor inicial para la unidad.
        símismo.assertEqual(res, 'paj')

    def test_trad_unidad_errónea_sin_especificar_leng(símismo):
        """
        Comprobar que devolvamos aviso error para unidades que no existen.
        """

        with símismo.assertRaises(ValueError):
            trads.trad_unid('¡Yo no existo!', leng_final='fr', falla_silencio=False)

    def test_agregar_sinónimo(símismo):
        """
        Comprobar que podemos agregar sinónimos de unidades.
        """

        trads.agregar_sinónimos(unid='foot', sinónimos='feet', leng='en', guardar=True)
        símismo.assertEqual(convertir(de='foot', a='feet'), 1)

    def test_agregar_sinónimo_listas(símismo):
        """
        Comprobar que podemos agregar sinónimos de unidades en formato de lista.
        """

        trads.agregar_sinónimos(unid='meter', sinónimos=['metre', 'm'], leng='en', guardar=False)
        símismo.assertEqual(convertir(de='metre', a='m'), 1)

    def test_agregar_trad(símismo):
        """
        Comprobar que podemos agregar traducciones de unidades.
        """

        trads.agregar_trad(unid='second', trad='پل', leng_orig='en', leng_trad='ur', guardar=False)
        símismo.assertEqual(convertir(de='second', a='پل'), 1)

    def test_agregar_trad_sin_especificar_leng(símismo):
        """
        Comprobar que podemos agregar traducciones de unidades sin especificar la lengua original de la unidad.
        """

        trads.agregar_trad(unid='second', trad='پل', leng_trad='ur', guardar=False)
        símismo.assertEqual(convertir(de='second', a='پل'), 1)

    def test_traducir_sinónimo_pint(símismo):
        """
        Comprobar que podemos traducir una unidad según su sinónimo Pint.
        """
        trads.agregar_trad('inch', trad='pulgado', leng_trad='es', guardar=False)
        símismo.assertEqual(trads.trad_unid('in', 'es'), 'pulgado')

    def test_agregar_trad_por_sinónimo_pint(símismo):
        """
        Comprobar que podemos agregar la traducción de una unidad según su sinónimo Pint.
        """

        trads.agregar_trad(unid='l', trad='litre', leng_trad='fr', guardar=False)
        trads.agregar_trad(unid='l', trad='litro', leng_trad='es', guardar=False)

        símismo.assertEqual(convertir('litres', a='litros'), 1)

    def test_agregar_trad_errónea(símismo):
        """
        Comprobar que tenemos error al agregar traducciones para unidades que no existen en la lengua especificada.
        """

        with símismo.assertRaises(ValueError):
            # "Year" no existe en castellano.
            trads.agregar_trad(unid='year', trad="année", leng_trad='fr', leng_orig='es')

    def test_agregar_trad_errónea_sin_especificar_leng(símismo):
        """
        Comprobar que tenemos error al agregar traducciones para unidades que no existen en cualquier idioma.
        """

        with símismo.assertRaises(ValueError):
            trads.agregar_trad(unid='¡Yo no existo!', trad="Je n'existe pas !", leng_trad='fr')



from tinamit.config import _
from .res import ResultadosSimul
from .var import Variable


class VariablesMod(object):

    def __init__(símismo, variables):
        símismo.variables = {v.nombre: v for v in variables}
        if len(símismo.variables) != len(variables):
            raise ValueError(_('Los variables de un modelo deben todos tener nombre distinto.'))

    def ingresos(símismo):
        return [v for v in símismo if v.ingr]

    def egresos(símismo):
        return [v for v in símismo if v.egr]

    def cambiar_vals(símismo, valores):
        for vr, vl in valores.items():
            símismo[vr].poner_val(vl)

    def reinic(símismo):
        for v in símismo:
            v.reinic()

    def gen_res(símismo, nombre, t, vars_interés):
        return ResultadosSimul(nombre=nombre, t=t, vars_interés=símismo._resolver_vars(vars_interés))

    def _resolver_vars(símismo, l_vars):
        if isinstance(l_vars, (Variable, str)):
            l_vars = [l_vars]
        return [símismo[v] for v in l_vars] if l_vars is not None else list(símismo)

    def __getitem__(símismo, itema):
        if itema in símismo:
            return símismo.variables[str(itema)]
        raise KeyError(itema)

    def __setitem__(símismo, llave, valor):
        símismo.variables[llave] = valor

    def __iter__(símismo):
        for v in símismo.variables.values():
            yield v

    def __contains__(símismo, itema):
        if isinstance(itema, str):
            return itema in símismo.variables
        else:
            return itema in símismo.variables.values()



import inspect
import os
import sys
import traceback
from importlib import import_module as importar_mod
from warnings import warn as avisar

from tinamit.config import _
from ._envolt import ModeloBF


def gen_bf(mod):
    """
    Genera un modelo BF desde un archivo o una subclase de ModeloBF.

    Parameters
    ----------
    mod: str or type or ModeloBF

    Returns
    -------
    ModeloBF
        Una instancia del modelo BF.
    """

    if isinstance(mod, ModeloBF):
        return mod
    elif callable(mod):
        if issubclass(mod, ModeloBF):
            return mod()
    elif isinstance(mod, str):
        return _extraer_de_archivo(mod)
    raise TypeError(
        _('Debes dar o una instancia o subclase de `ModeloDS`, o la dirección de un archivo que contiene una.')
    )


def _extraer_de_archivo(archivo):
    if not os.path.isfile(archivo):
        raise FileNotFoundError(_('El archivo siguiente no existe... :(\n\t{}').format(archivo))

    if os.path.splitext(archivo)[1] != '.py':
        raise ValueError(_('El archivo siguiente no parece ser un archivo Python.').format(archivo))

    dir_mod, nombre_mod = os.path.split(archivo)
    sys.path.append(dir_mod)
    módulo = importar_mod(os.path.splitext(nombre_mod)[0])
    instancias = {nmb: cls for nmb, cls in inspect.getmembers(módulo, lambda x: isinstance(x, ModeloBF))}
    clases = {
        nmb: cls for nmb, cls in inspect.getmembers(
            módulo, lambda x: (inspect.isclass(x) and issubclass(x, ModeloBF))
        )
    }
    if instancias:
        if 'Envoltura' in instancias:
            return instancias['Envoltura']
        return list(instancias.values())[0]

    potenciales = {}
    errores = {}
    for nmb, cls in clases.items():
        # noinspection PyBroadException
        try:
            potenciales[nmb] = cls()
        except NotImplementedError:
            continue
        except Exception:
            errores[nmb] = traceback.format_exc()
            continue

    if len(potenciales) == 1:
        return list(potenciales.values())[0]
    elif 'Envoltura' in potenciales:
        return potenciales['Envoltura']
    elif potenciales:
        nmb_elegida = list(potenciales)[0]
        avisar(_('\nHabía más de una instancia de "ModeloBF" en el fuente'
                 '\n\t{}'
                 '\n...y ninguna se llamaba "Envoltura". Tomaremos "{}" como la envoltura'
                 '\ny esperaremos que funcione. Si no te parece, asegúrate que la definición de clase o el'
                 '\nobjeto correcto se llame "Envoltura".'
                 ).format(archivo, nmb_elegida))
        return potenciales[nmb_elegida]

    raise AttributeError(_(
        'El archivo especificado: \n\t{}\nno contiene subclase o instancia de "EnvolturaBF" utilizable. '
        '\nErrores encontrados:{}'
    ).format(archivo, ''.join(['\n\n\t{}: \n{}'.format(nmb, e) for nmb, e in errores.items()])))



import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta

from tinamit.config import _
from tinamit.cositas import _gen_fecha
from tinamit.unids.conv import convertir


class Tiempo(object):
    """
    La base temporal para cada simulación.
    """
    def __init__(símismo, t, unid_paso):
        """

        Parameters
        ----------
        t: EspecTiempo
        unid_paso: str
            La unidad de tiempo.
        """

        símismo.tmñ_paso = t.tmñ_paso
        símismo.n_pasos = t.n_pasos
        símismo.f_inic = t.f_inic
        símismo.guardar_cada = t.guardar_cada

        símismo.unid_paso, símismo.fact_conv = a_unid_tnmt(unid_paso)

        símismo._convs = {}

        símismo.í = 0

    def t_guardar(símismo):
        """
        Si estamos en un paso del cual guardaremos los resultados.
        Returns
        -------
        bool
        """
        return not (símismo.í * símismo.tmñ_paso % símismo.guardar_cada)

    def pasos_avanzados(símismo, unid):
        """
        El número de pasos avanzados desde el último incremento.

        Parameters
        ----------
        unid: str
            La unidad en la cual querremos la respuesta.

        Returns
        -------
        int
        """
        return símismo._obt_fact_conv(unid) * símismo.fact_conv * símismo.tmñ_paso

    def fecha(símismo):
        pass

    def avanzar(símismo):
        """
        Avanzar el tiempo, si no hemos llegado ya al final de la simulación.
        Returns
        -------
        bool
            ``False`` si llegamos al final de la simulación, ``True`` si no.

        """
        if símismo.í < símismo.n_pasos:
            símismo.í += 1
            return símismo.í
        return False

    def eje(símismo):
        """
        El eje temporal.

        Returns
        -------
        pd.Index
        """
        return pd.Index(np.arange(
            símismo.n_pasos * símismo.tmñ_paso + 1,
            step=símismo.guardar_cada * símismo.tmñ_paso
        ))

    def _obt_fact_conv(símismo, unid):

        # Evitar buscar la misma unidad 2 veces porque la búsqueda puede ser lenta.
        try:
            return símismo._convs[unid]
        except KeyError:
            fact = convertir(de=símismo.unid_paso, a=unid, val=1)
            símismo._convs[unid] = round(fact)
            return fact

    def __len__(símismo):
        return símismo.n_pasos // símismo.guardar_cada + 1


class TiempoCalendario(Tiempo):
    def fecha(símismo):
        """
        La fecha actual.

        Returns
        -------
        datetime.datetime

        """
        unid_ft = a_unid_ft[símismo.unid_paso]
        return símismo.f_inic + relativedelta(**{unid_ft: símismo.í * símismo.fact_conv})

    def delta_relativo(símismo, n_pasos):
        """
        Devuelve un objeto de diferenci de tiempo para el número de pasos especificados.

        Parameters
        ----------
        n_pasos: int
            El número de pasos para avanzar.

        Returns
        -------
        relativedelta

        """
        if símismo.unid_paso in ['año', 'mes']:
            return relativedelta(**{a_unid_ft[símismo.unid_paso]: n_pasos * símismo.tmñ_paso * símismo.fact_conv})

        n_días = convertir(símismo.unid_paso, a='día', val=n_pasos * símismo.tmñ_paso * símismo.fact_conv)
        return relativedelta(days=n_días)

    def fecha_próxima(símismo):
        """
        La próxima fecha que vendrá.

        Returns
        -------
        datetime.datetime
        """
        return símismo.fecha() + símismo.delta_relativo(n_pasos=1)

    def eje(símismo):
        paso = símismo.fact_conv * símismo.guardar_cada * símismo.tmñ_paso
        if símismo.unid_paso == 'mes':
            eje = pd.to_datetime([
                símismo.f_inic + relativedelta(months=i)
                for i in range(0, símismo.n_pasos // símismo.guardar_cada + 1, paso)
            ])
            return eje
        return pd.date_range(
            símismo.f_inic, periods=símismo.n_pasos // símismo.guardar_cada + 1,
            freq=str(paso) + _a_unid_pandas[símismo.unid_paso]
        )

    def pasos_avanzados(símismo, unid):
        unid, fact = a_unid_tnmt(unid)
        if símismo.unid_paso == 'año' and unid == 'mes':
            return 12 * símismo.tmñ_paso * símismo.fact_conv / fact

        if símismo.unid_paso in ['año', 'mes']:
            delta = relativedelta(**{a_unid_ft[símismo.unid_paso]: símismo.tmñ_paso * símismo.fact_conv / fact})
            n_días = (símismo.fecha() - (símismo.fecha() - delta)).days
            return round(convertir('día', a=unid, val=n_días))

        return round(convertir(símismo.unid_paso, a=unid, val=símismo.tmñ_paso * símismo.fact_conv / fact))


class EspecTiempo(object):
    """
    Objeto para especificar opciones de tiempo para la simulación.
    """
    def __init__(símismo, n_pasos, f_inic=None, tmñ_paso=1, guardar_cada=1):
        """

        Parameters
        ----------
        n_pasos: int
            El número de pasos en la simulación.
        f_inic: str or datetime.datetime or datetime.date
            La fecha inicial, si hay.
        tmñ_paso: int
            El tamaño de cada paso.
        guardar_cada:
            La frequencia para guardar resultados.
        """

        if (int(tmñ_paso) != tmñ_paso) or (tmñ_paso < 1):
            raise ValueError(_('`tmñ_paso` debe ser un número entero superior a 0.'))
        símismo.f_inic = _gen_fecha(f_inic)

        símismo.tmñ_paso = tmñ_paso
        símismo.n_pasos = n_pasos
        símismo.guardar_cada = guardar_cada

    def gen_tiempo(símismo, unid_paso):
        """
        Genera un objeto :class:`Tiempo` con las especificaciones.

        Parameters
        ----------
        unid_paso: str
            La unidad de paso de la simulación.

        Returns
        -------
        Tiempo

        """
        if símismo.f_inic:
            return TiempoCalendario(t=símismo, unid_paso=unid_paso)
        else:
            return Tiempo(t=símismo, unid_paso=unid_paso)


def a_unid_tnmt(unid):
    """
    Convierte una unidad de tiempo a una de las reconocidas por Tinamït.

    Parameters
    ----------
    unid: str

    Returns
    -------
    tuple[str, float]
        La unidad Tinamït y el factor de conversión.
    """

    unid = unid.lower()
    aceptables = [
        'año', 'mes', 'semana', 'día', 'hora', 'minuto',
        'secundo', 'microsecundo', 'millisecundo', 'nanosecundo'
    ]
    for u in aceptables:
        if unid == u:
            return unid, 1
        try:
            factor = convertir(de=unid, a=u)
            if int(factor) == factor:
                return u, factor
        except (ValueError, KeyError):
            pass
    raise ValueError(
        _('La unidad de tiempo "{}" no se pudo convertir a años, meses, días, horas, minutos o secundos.').format(unid))


a_unid_ft = {
    'año': 'years',
    'mes': 'months',
    'semana': 'weeks',
    'día': 'days',
    'hora': 'hours',
    'minuto': 'minutes',
    'secundo': 'seconds',
    'millisecundo': 'milliseconds',
    'microsecundo': 'microseconds',
}

_a_unid_pandas = {
    'año': 'A',
    'mes': 'M',
    'día': 'D',
    'hora': 'H',
    'minuto': 'min',
    'secundo': 'S',
    'millisecundo': 'ms',
    'microsecundo': 'us',
}



from unittest import TestCase

import numpy as np
import numpy.testing as npt
import pandas as pd

from pruebas.recursos.bf.variantes import EjDeterminado, EjBloques
from pruebas.recursos.mod.prueba_mod import ModeloPrueba
from tinamit.mod.clima import Clima
from tinamit.tiempo.tiempo import EspecTiempo
from تقدیر.ذرائع import جےسن


class TestClima(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.lluvia = np.random.random(366 + 365 + 365)
        cls.fechas = pd.date_range('2000-01-01', '2002-12-31')

        cls.clima = Clima(lat=31.569, long=74.355, elev=100, fuentes=جےسن(
            {'بارش': cls.lluvia, 'تاریخ': cls.fechas}, 31.569, 74.355, 100
        ))

    def test_diario(símismo):
        mod = ModeloPrueba(unid_tiempo='días')
        mod.conectar_var_clima('Vacío', 'بارش', conv=1, combin='total')

        res = mod.simular(EspecTiempo(100, f_inic=símismo.fechas[0]), clima=símismo.clima, vars_interés='Vacío')

        npt.assert_equal(res['Vacío'].vals[:, 0], símismo.lluvia[:101])

    def test_mensual(símismo):
        mod = ModeloPrueba(unid_tiempo='mes')
        mod.conectar_var_clima('Vacío', 'بارش', conv=1, combin='total')

        res = mod.simular(EspecTiempo(2, f_inic=símismo.fechas[0]), clima=símismo.clima, vars_interés='Vacío')

        ref = np.array([
            np.sum(x) for x in
            [símismo.lluvia[: 31], símismo.lluvia[31: 31 + 29], símismo.lluvia[31 + 29: 31 + 29 + 31]]
        ])
        npt.assert_equal(res['Vacío'].vals[:, 0], ref)

    def test_anual(símismo):
        mod = ModeloPrueba(unid_tiempo='año')
        mod.conectar_var_clima('Vacío', 'بارش', conv=1, combin='total')

        res = mod.simular(EspecTiempo(2, f_inic=símismo.fechas[0]), clima=símismo.clima, vars_interés='Vacío')

        ref = np.array([
            np.sum(x) for x in
            [símismo.lluvia[: 366], símismo.lluvia[366: 366 + 365], símismo.lluvia[366 + 365: 366 + 365 + 365]]
        ])
        npt.assert_equal(res['Vacío'].vals[:, 0], ref)


class TestClimaBFs(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.lluvia = np.random.random(366 + 365)
        cls.fechas = pd.date_range('2000-01-01', '2001-12-31')

        cls.clima = Clima(lat=31.569, long=74.355, elev=100, fuentes=جےسن(
            {'بارش': cls.lluvia, 'تاریخ': cls.fechas}, 31.569, 74.355, 100
        ))

    def test_deter(símismo):
        mod = EjDeterminado(tmñ_ciclo=30, unid_tiempo='días')
        mod.conectar_var_clima('ingreso_ciclo', 'بارش', conv=1, combin='total')
        mod.conectar_var_clima('ingreso_paso', 'بارش', conv=1, combin='total')

        ref_paso = símismo.lluvia[:360]
        ref_ciclo = np.repeat(np.array([
            np.sum(símismo.lluvia[x * 30:x * 30 + 30]) for x in range(12)
        ]), 30)

        pruebas = {
            'ingreso_ciclo': ref_ciclo,
            'ingreso_paso': ref_paso
        }

        res = mod.simular(EspecTiempo(30 * 12 - 1, f_inic=símismo.fechas[0]), clima=símismo.clima)

        for prb, ref in pruebas.items():
            if prb == 'ingreso_paso':
                continue
            with símismo.subTest(prb):
                npt.assert_equal(res[prb].vals[:, 0].values, ref)

    def test_bloques(símismo):
        tmñ_bloques = [4, 5, 3]
        mod = EjBloques(tmñ_bloques=tmñ_bloques, unid_tiempo='meses')
        mod.conectar_var_clima('ingreso_paso', 'بارش', conv=1, combin='total')
        mod.conectar_var_clima('ingreso_bloque', 'بارش', conv=1, combin='total')
        mod.conectar_var_clima('ingreso_ciclo', 'بارش', conv=1, combin='total')

        n_días = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        sum_cum = np.cumsum(np.append([0], n_días))
        ref_paso = np.array([
            np.sum(símismo.lluvia[sum_cum[x]:sum_cum[x + 1]]) for x in range(24)
        ])

        sum_cum_bloq = np.cumsum(np.append([0], np.tile(tmñ_bloques, 2)))
        ref_bloques = np.repeat(np.array([
            np.sum(símismo.lluvia[sum_cum[sum_cum_bloq[i]]: sum_cum[sum_cum_bloq[i + 1]]])
            for i in range(len(sum_cum_bloq) - 1)
        ]), np.tile(tmñ_bloques, 2))

        ref_ciclo = np.empty(24)
        ref_ciclo[:12] = np.sum(símismo.lluvia[:np.sum(n_días[:12])])
        ref_ciclo[12:] = np.sum(símismo.lluvia[np.sum(n_días[:12]):])

        pruebas = {
            'ingreso_ciclo': ref_ciclo,
            'ingreso_paso': ref_paso,
            'ingreso_bloque': ref_bloques
        }

        res = mod.simular(EspecTiempo(23, f_inic=símismo.fechas[0]), clima=símismo.clima)
        for prb, ref in pruebas.items():
            if prb in ['ingreso_paso', 'ingreso_bloque']:
                continue
            with símismo.subTest(prb):
                npt.assert_equal(res[prb].vals[:, 0].values, ref)



import datetime as ft
import pickle
from multiprocessing import Pool as Reserva
from warnings import warn as avisar

import numpy as np

from tinamit.config import _, conf_mods
from .corrida import Corrida, Rebanada
from .extern import gen_extern
from .res import ResultadosGrupo, ResultadosSimul
from .vars_mod import VariablesMod
from ..tiempo.tiempo import EspecTiempo


class Modelo(object):
    """
    Todas las cosas en Tinamit son instancias de `Modelo`, que sea un modelo de dinámicas de los sistemas, un modelo de
    cultivos o de suelos o de clima, o un modelo conectado.
    Cada tipo de modelo se representa por subclases específicas. Por eso, la gran mayoría de los métodos definidos
    aquí se implementan de manera independiente en cada subclase de `Modelo`.
    """

    idioma_orig = 'es'  # Cambiar en tu modelo si el idioma de sus variables, etc. no es el castellano

    def __init__(símismo, variables, nombre):
        """
        La función de inicialización de todos modelos, conectados o no.

        Parameters
        ----------
        variables : VariablesMod
            Los variables del modelo.
        nombre : str
            El nombre del modelo.

        """

        símismo.nombre = nombre
        símismo.variables = variables
        símismo.corrida = None
        símismo.vars_clima = {}

    def unidad_tiempo(símismo):
        """
        Esta función debe devolver la unidad de tiempo empleada por el modelo.

        Returns
        -------
        str
            La unidad de tiempo (p. ej., 'meses', 'مہینہ', etc.)
        """

        raise NotImplementedError

    def simular(símismo, t, nombre='Tinamït', extern=None, clima=None, vars_interés=None):
        """
        
        Parameters
        ----------
        t: int or EspecTiempo
            La especificación del eje de tiempo. Si es ``int``, significará el número de pasos.
        nombre: str
            El nombre de la corrida.
        extern: Extern or pd.DataFrame or xr.Dataset or dict
            Valores externos para la simulación.
        clima: Clima
            El clima de la simulación.
        vars_interés: list
            Los variables para incluir en los resultados

        Returns
        -------
        ResultadosSimul

        """

        t = t if isinstance(t, EspecTiempo) else EspecTiempo(t)

        corrida = Corrida(
            nombre, t=t.gen_tiempo(símismo.unidad_tiempo()),
            extern=gen_extern(extern),
            vars_mod=símismo.variables,
            vars_interés=vars_interés,
            clima=clima
        )

        símismo.iniciar_modelo(corrida)
        símismo.correr()
        símismo.cerrar()

        return corrida.resultados

    def simular_grupo(símismo, ops_grupo, nombre='Tinamït', paralelo=False):
        """
        Efectua un grupo de simulaciones. Muy útil para accelerar corridas múltiples.

        Parameters
        ----------
        ops_grupo: PlantillaOpsSimulGrupo
            Las opciones de simulación en grupo.
        nombre: str
            El nombre de la simulación.
        paralelo: bool
            Si se simula en paralelo o no. Si el modelo no soporte corridas en paralelo, se ignorará este argumento.

        Returns
        -------
        ResultadosGrupo
        """

        if paralelo and not símismo.paralelizable():
            avisar(_(
                '\nEl modelo no se identifica como paralelizable. Para evitar el riesgo'
                '\nde errores de paralelización, correremos las corridas como simulaciones secuenciales '
                '\nnormales. '
                '\nSi tu modelo sí es paralelizable, crear un método nombrado `.paralelizable()` '
                '\nque devuelve ``True`` en tu clase de modelo para activar la paralelización.'
            ))
            paralelo = False

        res_grupo = ResultadosGrupo(nombre)
        if paralelo:
            l_trabajos = []
            copia_mod = pickle.dumps(símismo)  # Una copia de este modelo.

            # Crear la lista de información necesaria para las simulaciones en paralelo...
            for ops in ops_grupo:
                l_trabajos.append((copia_mod, ops))

            # Hacer las corridas en paralelo
            with Reserva() as r:
                res_paralelo = r.map(_correr_modelo, l_trabajos)
            for res in res_paralelo:
                res_grupo[str(res)] = res

        else:
            for ops in ops_grupo:
                res = símismo.simular(**ops)
                res_grupo[str(res)] = res

        return res_grupo

    def iniciar_modelo(símismo, corrida):
        """
        Inicia la simulación. En general no llamarías esta función directamente.
        
        No se te olvide una llamada al ``super`` cuando reimplementas esta función.
        
        Parameters
        ----------
        corrida: Corrida
            La corrida.

        """
        símismo.corrida = corrida
        corrida.variables.reinic()

        if corrida.extern:
            símismo.cambiar_vals(
                {vr: vl for vr, vl in corrida.obt_extern_act().items() if vr in símismo.variables}
            )
        if corrida.clima and símismo.vars_clima:
            t = símismo.corrida.t
            corrida.clima.inicializar(t)
            símismo._act_vals_clima(
                t.fecha(), t.fecha_próxima() - 1 * t.fecha().freq
            )
        corrida.actualizar_res()

    def correr(símismo):
        """
        Efectuar una simulación ya inicializada. En general, no llamarías esta función directamente.
        """

        intento = símismo._correr_hasta_final()
        if intento is not None:
            símismo.corrida.resultados.poner_vals_t(intento)
        else:
            while símismo.corrida.t.avanzar():
                símismo.incrementar(
                    Rebanada(
                        símismo.corrida.t.pasos_avanzados(símismo.unidad_tiempo()),
                        resultados=símismo.corrida.resultados
                    )
                )
                símismo.corrida.actualizar_res()

    def _correr_hasta_final(símismo):
        """
        Implementar si tu modelo puede correr de una vez hasta el final de la simulación, y después devolver los 
        resultados a cada paso de tiempo. Sirve a accelerar simulaciones no conectadas.
        
        Returns
        -------
        dict
            Un diccionario con matrices de los variables de egreso. Eje 0 debe ser el eje de tiempo.

        """
        return None

    def incrementar(símismo, rebanada):
        """
        Incrementa el modelo. En general, no llamarías esta función directamente.
        
        No se te olvide una llamada al ``super`` cuando reimplementas esta función.
        
        Parameters
        ----------
        rebanada: Rebanada
            La rebanada del incremento.

        """
        if símismo.corrida.extern:
            símismo.cambiar_vals(símismo.corrida.obt_extern_act())

        if símismo.corrida.clima and símismo.vars_clima:
            t = símismo.corrida.t
            símismo._act_vals_clima(t.fecha(), t.fecha_próxima() - 1 * t.fecha().freq)

    def cerrar(símismo):
        """
        Esta función toma acciones necesarias para terminar la simulación y cerrar el modelo, si aplica.
        """
        pass

    def cambiar_vals(símismo, valores):
        """
        Esta función cambia el valor de uno o más variables del modelo.

        Parameters
        ----------
        valores : dict
            Un diccionario de variables y sus valores para cambiar.
        """

        # Cambia primero el valor en el diccionario interno del Modelo
        símismo.variables.cambiar_vals(valores=valores)

    @classmethod
    def obt_conf(cls, llave, auto=None, cond=None, mnsj_err=None):
        """
        Obtiene un valor de configuración de la subclase de modelo.
        
        Parameters
        ----------
        llave: str
            El parámetro de configuración.
        auto: str or int or float or list or bool or dict
            Un valor automático a aplicar si no se encuentra en el diccionario de configuración.
        cond:
            Una condición para validar el valor; si no pasa la condición, se tratará como valor que falta.
        mnsj_err:
            Un mensaje de aviso para devolver al usuario si no se encuentra el valor.

        Returns
        -------
        str, int, float, list, bool, dict
            El valor de configuración
        """

        auto = auto or []
        if isinstance(auto, str):
            auto = [auto]

        try:
            op = conf_mods[cls.__name__][llave]
            if cond is None or cond(op):
                return op
        except KeyError:
            pass

        for op in auto:
            if cond is None or cond(op):
                conf_mods[cls.__name__, llave] = op
                return op

        if mnsj_err:
            avisar(mnsj_err)

    @classmethod
    def estab_conf(cls, llave, valor):
        """
        Establece un valor de configuración.
        
        Parameters
        ----------
        llave: str
            El parámetro de configuración.
        valor: str or int or float or list or bool or dict
            El valor del parámetro.

        """
        conf_mods[cls.__name__, llave] = valor

    @classmethod
    def instalado(cls):
        """
        Si tu modelo depiende en una instalación de otro programa externo a Tinamït, puedes reimplementar esta función
        para devolver ``True`` si el modelo está instalado y ``False`` sino.

        Returns
        -------
        bool
            Si el modelo está instalado completamente o no.
        """

        return True

    def paralelizable(símismo):
        """
        Indica si el modelo actual se puede paralelizar de manera segura o no. Si implementas una subclase
        paralelizable, reimplementar esta función para devolver ``True``.
        
        ¿No sabes si es paralelizable tu modelo?
        
        **Respuesta larga**: Si el modelo se puede paralelizar (con corridas de nombres distintos) sin encontrar 
        dificultades técnicas (sin riesgo que las corridas paralelas terminen escribiendo en los mismos archivos de 
        egreso), entonces sí es paralelizable tu modelo.
        
        **Respuesta rápida**: 95% seguro que sí.
        
        Returns
        -------
        bool:
            Si el modelo es paralelizable o no.
        """

        return False

    def conectar_var_clima(símismo, var, var_clima, conv, combin='prom'):
        """
        Conecta un variable climático.

        Parameters
        ----------
        var : str
            El nombre interno del variable en el modelo.
        var_clima : str
            El nombre oficial del variable climático.
        conv : number
            La conversión entre el variable clima en Tinamït y el variable correspondiente en el modelo.
        combin : str or function or None
            Si este variable se debe adicionar o tomar el promedio entre varios pasos. Puede ser ``prom``, ``total``,
            o una función. Si es ``None``, se tomará el último día en el caso de pasos de más de 1 día.
        """

        combins = {
            'prom': np.mean,
            'total': np.sum
        }
        if isinstance(combin, str):
            combin = combins[combin.lower()]

        símismo.vars_clima[var] = {
            'nombre_tqdr': var_clima,
            'combin': combin,
            'conv': conv
        }

    def _act_vals_clima(símismo, f_0, f_1, vars_clima=None):
        """
        Actualiza los variables climáticos. Esta función es la automática para cada modelo. Si necesitas algo más
        complicado (como, por ejemplo, predicciones por estación), la puedes cambiar en tu subclase.

        Parameters
        ----------
        f_0 : ft.date | ft.datetime
            La fecha actual.
        f_1 : ft.date | ft.datetime
            La próxima fecha.
        vars_clima : dict
        """
        vars_clima = vars_clima or símismo.vars_clima
        datos = símismo.corrida.clima.combin_datos(vars_clima=vars_clima, f_inic=f_0, f_final=f_1)

        símismo.cambiar_vals(valores=datos)

    def __str__(símismo):
        return símismo.nombre


def _correr_modelo(x):
    """
    Función para inicializar y correr un modelo en paralelo.

    Parameters
    ----------
    x : tuple[Modelo, dict]
        Los parámetros. El primero es el modelo, el segundo el diccionario de parámetros para la simulación.

    Returns
    -------
    dict
        Los resultados de la simulación.
    """

    estado_mod, d_args = x

    mod = pickle.loads(estado_mod)

    # Después, simular el modelo y devolver los resultados, si hay.
    return mod.simular(**d_args)



import os

from tinamit.config import _
from .pysd import ModeloPySD
from .vensim_dll import ModeloVensimDLL
from ._envolt import ModeloDS

_subclases = [ModeloPySD, ModeloVensimDLL]


def registrar_envolt_mds(envoltura):
    """
    Registra una nueva envoltura en Tinamït.

    Parameters
    ----------
    envoltura:
        La nueva envoltura.

    """
    _subclases.insert(0, envoltura)


def olvidar_envolt_mds(envoltura):
    """
    Borra una envoltura del registro global.

    Parameters
    ----------
    envoltura:
        La envoltura que ya no quieres.

    """
    _subclases.remove(envoltura)


class ErrorNoInstalado(OSError):
    """
    Error para devolver si no está instalada una envoltura.
    """
    pass


def gen_mds(archivo):
    """
    Automáticamente generar un :class:`~tinamit.envolt.mds._envolt.ModeloDS` desde un archivo.

    Parameters
    ----------
    archivo: str

    Returns
    -------
    ModeloDS

    """
    ext = os.path.splitext(archivo)[1]

    instaladas = [s for s in _subclases if s.instalado()]
    for cls in instaladas:
        ext_cls = [cls.ext] if isinstance(cls.ext, str) else cls.ext
        for e in ext_cls:
            e = '.' + e if ext[0] != '.' else e
            if ext == e:
                return cls(archivo)
    raise ErrorNoInstalado(_('No se encontró envoltura instalada para archivos de tipo "{}".'.format(ext)))



import math as mat

from dateutil.relativedelta import relativedelta as deltarelativo
from tinamit.tiempo.tiempo import a_unid_tnmt, a_unid_ft

from ._impac import ModeloImpaciente, VariablesModImpaciente, VarPaso


class ModeloDeterminado(ModeloImpaciente):
    """
    La clase pariente para todos modelos que correr por un número predeterminado de pasos a cada simulación.
    """

    def incrementar(símismo, rebanada):
        # Para simplificar el código un poco.
        p = símismo.paso_en_ciclo
        n_pasos = rebanada.n_pasos

        # Aplicar el incremento de paso
        p_después = (p + n_pasos) % símismo.tmñ_ciclo

        # Guardar el pasito actual para la próxima vez.
        símismo.paso_en_ciclo = p_después

        # Si hay que avanzar el modelo externo, lanzar una su simulación aquí.
        if p_después == 0 or ((p + n_pasos) >= símismo.tmñ_ciclo):
            # El número de ciclos para simular
            c = mat.ceil(n_pasos / símismo.tmñ_ciclo)  # type: int

            # Avanzar la simulación
            símismo.avanzar_modelo(n_ciclos=c)

        # Actualizar el paso en los variables
        símismo.variables.act_paso(símismo.paso_en_ciclo)

        super().incrementar(rebanada)

    def _act_vals_clima(símismo, f_0, f_1, vars_clima=None):

        vars_clima = vars_clima or símismo.vars_clima
        # Actualizar datos de clima
        p = símismo.paso_en_ciclo

        if símismo.corrida.clima and vars_clima and p == (símismo.tmñ_ciclo - 1):
            t = símismo.corrida.t
            f_inic = t.fecha()

            un_día = deltarelativo(days=1)

            vars_clim_paso = {
                vr: d for vr, d in vars_clima.items() if isinstance(símismo.variables[vr], VarPasoDeter)
            }
            vars_clim_ciclo = {vr: d for vr, d in vars_clima.items() if vr not in vars_clim_paso}

            base_t, factor = a_unid_tnmt(símismo.unidad_tiempo())

            f_final = f_inic + deltarelativo(**{a_unid_ft[base_t]: factor * símismo.tmñ_ciclo}) - un_día

            datos_ciclo = símismo.corrida.clima.combin_datos(
                vars_clima=vars_clim_ciclo, f_inic=f_inic, f_final=f_final
            )
            for var, vl in datos_ciclo.items():
                símismo.variables[var].poner_val(vl)

            for i in range(símismo.tmñ_ciclo):

                f_final = f_inic + deltarelativo(**{a_unid_ft[base_t]: factor}) - un_día

                # Calcular los datos
                datos = símismo.corrida.clima.combin_datos(
                    vars_clima=vars_clim_paso, f_inic=f_inic, f_final=f_final
                )

                # Aplicar los valores de variables calculados
                for var, datos_vrs in datos.items():
                    # Guardar el valor para esta estación
                    símismo.variables[var].poner_vals_paso(datos_vrs, paso=i)

                f_inic = f_final + un_día

    def avanzar_modelo(símismo, n_ciclos):
        """
        Avanzar el modelo por un número determinado de ciclos.

        Parameters
        ----------
        n_ciclos: int
            El número de ciclos que hay que avanzar.

        """
        raise NotImplementedError

    def unidad_tiempo(símismo):
        raise NotImplementedError


class VariablesModDeter(VariablesModImpaciente):
    """
    Representa los variables de un modelo :class:`~tinamit.envolt.bf._deter.Determinado`.
    """
    pass


class VarPasoDeter(VarPaso):
    """
    Un variable de un modelo :class:`~tinamit.envolt.bf._deter.Determinado` que toma un valor distinto a cada
    paso (y no solamente a cada ciclo de simulación).
    """

    def __init__(símismo, nombre, unid, ingr, egr, tmñ_ciclo, inic=0, líms=None, info=''):
        super().__init__(nombre, unid, ingr, egr, tmñ_ciclo=tmñ_ciclo, inic=inic, líms=líms, info=info)



from pkg_resources import resource_filename

from .config import conf, conf_mods, trads

# Cosas básicas
__autor__ = 'Julien Malard'

__correo__ = 'Contacto: julien.malard@mail.mcgill.ca'

with open(resource_filename('tinamit', 'versión.txt')) as _archivo_versión:
    __versión__ = _archivo_versión.read().strip()

__version__ = __versión__



import ctypes
import os
import sys

import numpy as np

from tinamit.config import _
from tinamit.mod import Variable
from . import _funcs as f
from ._vars import VarAuxEditable, VariablesModVensimDLL
from .._envolt import ModeloDS
from .._vars import VarNivel, VarConstante, VarInic, VarAuxiliar


class ModeloVensimDLL(ModeloDS):
    """
    Esta es la envoltura para modelos de tipo Vensim. Puede leer y controlar cualquier modelo Vensim para que
    se pueda emplear en Tinamït.
    Necesitarás la versión DSS de Vensim para que funcione.
    """
    ext = ['.vpm']

    def __init__(símismo, archivo, nombre='mds'):
        símismo.inicializado = False
        símismo.mod = f.cargar_mod_vensim(_obt_dll_vensim(), archivo)
        símismo.paso = f.obt_paso_inicial(símismo.mod)
        super().__init__(_gen_vars(símismo.mod), nombre)

    def unidad_tiempo(símismo):
        return f.obt_unid_tiempo(símismo.mod)

    def iniciar_modelo(símismo, corrida):
        símismo.corrida = corrida
        corrida.variables.reinic()

        t = corrida.t
        # En Vensim, tenemos que incializar los valores de variables no editables antes de empezar la simulación.
        no_editables = [str(v) for v in símismo.variables.no_editables()]
        if corrida.extern:
            símismo.cambiar_vals(
                {vr: vl.values for vr, vl in corrida.obt_extern_act().items() if vr in no_editables}
            )
        if corrida.clima:
            símismo.cambiar_vals(
                {vr: vl.values for vr, vl in corrida.clima.obt_todos_vals(t, símismo.vars_clima).items()}
            )

        f.inic_modelo(símismo.mod, paso=t.tmñ_paso, n_pasos=t.n_pasos, nombre_corrida=str(corrida))

        # Aplicar los valores iniciales de variables editables
        if corrida.extern:
            símismo.cambiar_vals(
                {var: val for var, val in
                 corrida.obt_extern_act(var=[str(v) for v in símismo.variables.editables()]).items()}
            )

        # Debe venir después de `f.inic_modelo()` sino no obtenemos datos para los variables
        símismo._leer_vals_de_vensim()

        símismo.inicializado = True

        corrida.actualizar_res()
        # NO llamamos a super() porque ya hicimos todo aquí. Se tenía que reimplementar para poder separar
        # la actualización de variables dinámicos y constantes en Vensim DLL.

    def incrementar(símismo, rebanada):
        corrida = símismo.corrida
        # Establecer el paso.
        paso = corrida.t.tmñ_paso
        if paso != símismo.paso:
            f.estab_paso(símismo.mod, paso)
            símismo.paso = paso

        f.avanzar_modelo(símismo.mod)
        símismo._leer_vals_de_vensim(rebanada.resultados.variables())

        super().incrementar(rebanada)

    def cambiar_vals(símismo, valores):
        super().cambiar_vals(valores)

        for var, val in valores.items():
            # Para cada variable para cambiar...

            if símismo.variables[var].dims == (1,):
                # Si el variable no tiene dimensiones (subscriptos)...

                # Actualizar el valor en el modelo Vensim.
                if not np.isnan(val):
                    f.cambiar_val(símismo.mod, var, val)

            else:
                # Para hacer: opciones de dimensiones múltiples
                # La lista de subscriptos
                subs = símismo.variables[var].subs

                if isinstance(val, np.ndarray) and val.shape:
                    matr = val
                else:
                    matr = np.empty(len(subs))
                    matr[:] = val

                for n, s in enumerate(subs):
                    var_s = var + s
                    val_s = matr[n]
                    if not np.isnan(val_s):
                        f.cambiar_val(símismo.mod, var_s, val_s)

    def cerrar(símismo):
        """
        Cierre la simulación Vensim.
        """

        f.cerrar_vensim(símismo.mod)

        #
        f.vdf_a_csv(símismo.mod, símismo.corrida.nombre)

    @classmethod
    def instalado(cls):
        return _obt_arch_dll_vensim() is not None

    def paralelizable(símismo):
        return True

    def _leer_vals_de_vensim(símismo, l_vars=None):
        if isinstance(l_vars, Variable):
            l_vars = [l_vars]
        elif l_vars is None:
            l_vars = símismo.variables

        for v in l_vars:
            val = f.obt_val_var(símismo.mod, str(v), v.subs)

            # Guardar en el diccionario interno.
            símismo.variables.cambiar_vals({v: val})


def _obt_dll_vensim():
    if sys.platform[:3] != 'win':
        raise OSError(
            _('\nDesafortunadamente, el DLL de Vensim funciona únicamente en Windows.'
              '\nPuedes intentar la envoltura ModeloPySD con un modelo .mdl en vez.')
        )
    return ctypes.WinDLL(_obt_arch_dll_vensim())


def _obt_arch_dll_vensim():
    return ModeloVensimDLL.obt_conf(
        'dll', auto=[
            'C:/Windows/System32/vendll32.dll',
            'C:/Windows/SysWOW64/vendll32.dll'
        ], cond=os.path.isfile,
        mnsj_err=_(
            '\nDebes instalar Vensim DSS en tu computadora y, si todavía aparece este mensaje,'
            '\nespecificar la ubicación del dll manualmente, p. ej.'
            '\n\tModeloVensimDLL.estab_conf("dll", "C:/Camino/raro/para/vendll32.dll")'
            '\npara poder hacer simulaciones con Vensim DSS.'
        )
    )


def _gen_vars(mod):
    l_vars = []

    vars_y_tipos = {v: f.obt_atrib_var(mod, v, cód_attrib=14).lower() for v in f.obt_vars(mod)}

    editables = f.obt_editables(mod)

    omitir = [
        'Data', 'Constraint', 'Lookup', 'Group', 'Subscript Range', 'Test Input', 'Time Base',
        'Subscript Constant'
    ]

    for var, tipo_var in vars_y_tipos.items():
        # Para cada variable...

        # No incluir los variables de verificación (pruebas de modelo) Vensim, de subscriptos, de datos, etc.
        if tipo_var in omitir:
            continue

        # Sacar los límites del variable
        líms = (f.obt_atrib_var(mod, var, cód_attrib=11), f.obt_atrib_var(mod, var, cód_attrib=12))
        líms = tuple(float(lm) if lm != '' else None for lm in líms)

        # Leer la descripción del variable.
        info = f.obt_atrib_var(mod, var, 2)

        # Sacar sus unidades
        unid = f.obt_atrib_var(mod, var, cód_attrib=1)

        # Sacar las dimensiones del variable
        subs = f.obt_atrib_var(mod, var, cód_attrib=9)
        if len(subs):
            dims = (len(subs),)  # Para hacer: permitir más de 1 dimensión
        else:
            dims = (1,)
            subs = None
        inic = np.zeros(dims)

        # Leer la ecuación del variable, sus hijos y sus parientes directamente de Vensim
        ec = f.obt_atrib_var(mod, var, 3)
        parientes = [v for v in f.obt_atrib_var(mod, var, 4) if v in vars_y_tipos]

        if tipo_var == 'constant' or (tipo_var == 'auxiliar' and not len(parientes)):
            cls = VarConstante
        elif tipo_var == 'level':
            cls = VarNivel
        elif tipo_var == 'auxiliary':
            cls = VarAuxEditable if var in editables else VarAuxiliar
        elif tipo_var == 'initial':
            cls = VarInic
        else:
            raise ValueError(tipo_var)

        l_vars.append(cls(var, unid=unid, ec=ec, subs=subs, parientes=parientes, inic=inic, líms=líms, info=info))

    return VariablesModVensimDLL(l_vars)



import os
from ast import literal_eval

import pysd
from tinamit.cositas import arch_más_recién


def gen_mod_pysd(archivo):
    nmbr, ext = os.path.splitext(archivo)

    if ext == '.py':
        return pysd.load(archivo)

    arch_py = nmbr + '.py'
    if os.path.isfile(arch_py) and arch_más_recién(arch_py, archivo):
        return pysd.load(nmbr + '.py')
    else:
        return pysd.read_vensim(archivo) if ext == '.mdl' else pysd.read_xmile(archivo)


def obt_paso_mod_pysd(mod):
    doc = mod.components.time_step.__doc__
    partes = doc.split()
    return decodar(partes[partes.index('Units:')+1])


def decodar(tx):
    if tx.startswith('b\''):
        return literal_eval(tx).decode('unicode_escape')
    return tx



import threading
from warnings import warn as avisar

import numpy as np

from tinamit.config import _
from tinamit.envolt.bf import gen_bf
from tinamit.envolt.mds import gen_mds
from tinamit.mod import Modelo
from tinamit.mod.corrida import Rebanada
from tinamit.unids.conv import convertir
from ._conex import Conex, ConexionesVars
from ._vars import VariablesConectado


class SuperConectado(Modelo):
    """
    Esta clase representa el más alto nivel posible de modelo conectado. Tiene la función muy útil de poder conectar
    instancias de sí misma, así permitiendo la conexión de números arbitrarios de modelos anidados.
    """

    def __init__(símismo, modelos, nombre='SuperConectado'):
        símismo.modelos = _ModelosConectados(modelos)
        símismo.conexiones = ConexionesVars()

        super().__init__(variables=VariablesConectado(símismo.modelos), nombre=nombre)

    def conectar_vars(símismo, var_fuente, var_recip, modelo_fuente=None, modelo_recip=None, conv=None):
        modelo_fuente = símismo._mod_de_var(var_fuente) if modelo_fuente is None else símismo.modelos[modelo_fuente]
        modelo_recip = símismo._mod_de_var(var_recip) if modelo_recip is None else símismo.modelos[modelo_recip]

        var_fuente = modelo_fuente.variables[var_fuente]
        var_recip = modelo_recip.variables[var_recip]

        símismo.conexiones.agregar(
            Conex(var_fuente, modelo_fuente, var_recip, modelo_recip, conv=conv)
        )

    def desconectar_vars(símismo, var_fuente, modelo_fuente, modelo_recip=None, var_recip=None):
        símismo.conexiones.quitar(var_fuente, modelo_fuente, modelo_recip, var_recip)

    def cerrar(símismo):
        for m in símismo.modelos:
            m.cerrar()

    def paralelizable(símismo):
        return all(m.paralelizable() for m in símismo.modelos)

    def unidad_tiempo(símismo):

        unids = list({m.unidad_tiempo() for m in símismo.modelos})
        factores_conv = [1]
        factores_conv[0] = 1
        for u in unids[1:]:
            # Intentar convertir la unidad del submodelo a la unidad de base
            factores_conv.append(convertir(de=unids[0], a=u))
        factores_conv = np.array(factores_conv)
        factores_conv = np.divide(factores_conv, factores_conv.min())

        facts_conv_ent = np.round(factores_conv)
        if not np.array_equal(facts_conv_ent, factores_conv):
            # Si todos los factores no son números enteros, tendremos que aproximar.
            avisar(
                _('Las unidades de tiempo de los modelos ({}) no tienen denominator común. Se aproximará la '
                  'conversión.').format(', '.join(unids))
            )

        return next(u for u, c in zip(unids, factores_conv) if c == 1)

    def iniciar_modelo(símismo, corrida):
        super().iniciar_modelo(corrida)

        for m in símismo.modelos:
            m.iniciar_modelo(corrida)

        símismo._intercambiar_vars()
        corrida.actualizar_res()

    def incrementar(símismo, rebanada):
        def incr_mod(mod, d, reb):
            try:
                mod.incrementar(reb)
            except BaseException as e:
                d[str(mod)] = e
                raise

        # Un diccionario para comunicar errores
        dic_err = {str(x): False for x in símismo.modelos}

        # Un hilo para cada modelo
        l_hilo = [threading.Thread(
            name='hilo_%s' % str(mod),
            target=incr_mod,
            args=(mod, dic_err, Rebanada(
                n_pasos=símismo.corrida.t.pasos_avanzados(mod.unidad_tiempo()),
                resultados=rebanada.resultados[str(mod)],
            )))
            for mod in símismo.modelos
        ]

        # Empezar los hilos al mismo tiempo
        for hilo in l_hilo:
            hilo.start()

        # Esperar que los dos hilos hayan terminado
        for hilo in l_hilo:
            hilo.join()

        # Verificar si hubo error
        for mod, err in dic_err.items():
            if err:
                raise ChildProcessError(_('Hubo error en el modelo "{m}": {e}').format(m=mod, e=err))

        símismo._intercambiar_vars()

    def cambiar_vals(símismo, valores):
        for mod in símismo.modelos:
            mod.cambiar_vals({vr: vl for vr, vl in valores.items() if vr in mod.variables})

    def _mod_de_var(símismo, var):
        return next(m for m in símismo.modelos if var in m.variables)

    def _intercambiar_vars(símismo):
        for c in símismo.conexiones:  # para hacer: por modelo para conexión más rápida
            val_fuente = c.var_fuente.obt_val()

            var_recip = c.modelo_recip.variables[c.var_recip]
            c.modelo_recip.cambiar_vals({str(var_recip): val_fuente * c.conv})


class Conectado(SuperConectado):
    """
    Un modelo que conecta un :class:`~tinamit.envolt.mds.ModeloDS` con un
    :class:`~tinamit.envolt.bf._envolt.ModeloBF`.
    """

    def __init__(símismo, bf, mds, nombre='conectado'):
        símismo.bf = gen_bf(bf)
        símismo.mds = gen_mds(mds) if isinstance(mds, str) else mds

        super().__init__([símismo.bf, símismo.mds], nombre=nombre)

    def conectar(símismo, var_mds, var_bf, mds_fuente, conv=None):
        """
        Una función para conectar variables entre el modelo biofísico y el modelo DS.

        Parameters
        ----------
        var_mds: str
            El nombre del variable en el modelo DS.
        var_bf: str
            El nombre del variable correspondiente en el modelo biofísico.
        mds_fuente: bool
            Si ``True``, el modelo DS es el modelo fuente para la conexión. Sino, será el modelo biofísico.
        conv: float
            El factor de conversión entre los variables.

        """

        # Establecer los variables y modelos fuentes y recipientes.
        modelo_fuente = símismo.mds if mds_fuente else símismo.bf
        var_fuente = var_mds if mds_fuente else var_bf

        modelo_recip = símismo.bf if mds_fuente else símismo.mds
        var_recip = var_bf if mds_fuente else var_mds

        # Llamar la función apropiada de la clase superior.
        símismo.conectar_vars(
            var_fuente=var_fuente, modelo_fuente=modelo_fuente,
            var_recip=var_recip, modelo_recip=modelo_recip, conv=conv
        )

    def desconectar(símismo, var_mds):
        """
        Esta función deshacer una conexión entre el modelo biofísico y el modelo DS. Se especifica la conexión por
        el nombre del variable en el modelo DS.

        Parameters
        ----------
        var_mds: str
            El nombre del variable conectado en el modelo DS.

        """

        # Llamar la función apropiada de la clase superior.
        símismo.desconectar_vars(var_fuente=var_mds, modelo_fuente=símismo.mds)


class _ModelosConectados(object):
    def __init__(símismo, modelos):
        símismo._modelos = {str(m): m for m in modelos}
        if len(símismo._modelos) != len(modelos):
            raise ValueError(_('Los modelos conectados deben todos tener nombres distintos.'))

    def __iter__(símismo):
        for m in símismo._modelos.values():
            yield m

    def __getitem__(símismo, itema):
        return símismo._modelos[str(itema)]



import os

from tinamit.Conectado import Conectado

modelo = Conectado()

directorio = os.path.dirname(__file__)

modelo.estab_mds(os.path.join(directorio, "mds_bosques.mdl"))
modelo.estab_bf(os.path.join(directorio, 'bf_bosques.py'))

modelo.conectar(var_ds='Lluvia', var_bf='Lluvia', mds_fuente=False)
modelo.conectar(var_ds='Bosques', var_bf='Bosques', mds_fuente=True)

res = modelo.simular(paso=1, t_final=100, nombre='Corrida_Tinamit')



import pint
import regex as re

from tinamit.config import _
from tinamit.unids import regu, C_, trad_unid
from tinamit.unids.trads import buscar_singular


def convertir(de, a, val=1, lengua=None):
    """
    Esta función convierte un valor de una unidad a otra.

    Parameters
    ----------
    de : str
        La unidad original.
    a : str
        La unidad final.
    val : float | int
        El valor para convertir.
    lengua : str
        La lengua en la cual las unidades están especificadas.

    Returns
    -------
    float
        El valor convertido.
    """

    # Guardar una copia de los unidades iniciales
    de_orig = de
    a_orig = a

    # Estar seguro que diferencias en mayúsculos o espacios no causen problemas
    de = de.lower().strip()
    a = a.lower().strip()

    # Si son las mismas unidades, no hay nada que hacer, por supuesto.
    if de == a:
        return val

    # Reformatear exponentes para que los entienda Pint
    de = re.sub(r'([\p{l}\p{m}]+\w*)(-?[\p{n}]+)', repl=r'\1 ^ \2', string=de)
    a = re.sub(r'([\p{l}\p{m}]+\w*)(-?[\p{n}]+)', repl=r'\1 ^ \2', string=a)

    # Leer los nombres de todas las unidades presentes.
    unids_pres_de = re.findall(r'[\p{l}\p{m}]+', de)
    unids_pres_a = re.findall(r'[\p{l}\p{m}]+', a)
    unids_pres = list(set(unids_pres_de + unids_pres_a))

    # Agregar unidades no reconocidas por Pint
    for u in unids_pres:
        # Para cada tipo_mod de unidad detectada en las unidades iniciales...

        try:
            # Ver si Pint lo reconoce
            regu.parse_units(u)

        except pint.UndefinedUnitError:
            # Si Pint no lo reconoce...

            try:
                # Primero intentamos traducirlo para Pint
                u_t = trad_unid(u, leng_final='en', leng_orig=lengua)

                if u_t != u:
                    # Si la traducción se efectuó...

                    # ...reemplazar las unidades iniciales con la versión traducida.
                    regu.parse_units(u_t)  # Verificar que la traducción sí funcione.
                    if u in unids_pres_de:
                        de = re.sub(r'%s(?![\p{l}\p{m}])' % u, repl=u_t, string=de)
                    if u in unids_pres_a:
                        a = re.sub(r'%s(?![\p{l}\p{m}])' % u, repl=u_t, string=a)

                else:
                    # Si no encontramos traducción, registramos la nueva unidad con Pint.
                    definir_en_regu(unid=u, base=buscar_singular(u)[-1])

            except pint.UndefinedUnitError:
                # Si Pint no reconoció aún la traducción, tenemos que registrar la unidad.
                definir_en_regu(unid=u, base=buscar_singular(u)[-1])

    # Intentar convertir
    try:
        cant = C_(val, de).to(a)
        cant_convtd = cant.magnitude  # Guardar el valor convertido

    except (pint.errors.DimensionalityError, pint.errors.UndefinedUnitError):
        # Si hasta ahora tenemos error, no puedo hacer nada más.
        raise ValueError(_('No se pudo convertir "{}" a "{}".').format(de_orig, a_orig))

    # Devolver la cantidad convertida
    return cant_convtd


def nueva_unidad(unid, ref, conv):
    regu.define('{unid} = {ref} * {conv}'.format(unid=unid, ref=ref, conv=conv))


def definir_en_regu(unid, base):
    """
    Esta funcion define una nueva unidad en el registro de unidades.

    Parameters
    ----------
    unid : str
        La unidad.
    base : str
        La dimensionalidad de la unidad para pasar a Pint.

    """

    # Detectar si la dimensionalidad ya se registró en Pint o no.
    try:
        base = regu.get_base_units(base)
        nueva_base = False
    except pint.errors.UndefinedUnitError:
        nueva_base = True

    if nueva_base:
        # Registrar la dimensionalidad, si necesario.
        regu.define('{b} = [{b}]'.format(b=base))
    if base != unid or not nueva_base:
        # Agregar la unidad a Pint si ya no se agregó con la dimensionalidad.
        regu.define('{u} = {b}'.format(u=unid, b=base))



import os

import numpy as np
import pandas as pd
import xarray as xr
from tinamit.config import _

from .fuente import FuentePandas, FuenteDic, FuenteVarXarray, FuenteBaseXarray, FuenteCSV, Fuente


class BD(object):
    """
    Una base de datos combina varias :class:`~tinamit.datos.fuente.Fuente`.
    """
    def __init__(símismo, fuentes):
        """

        Parameters
        ----------
        fuentes: Fuente or list
            Las fuentes de la base de datos.
        """
        fuentes = [fuentes] if not isinstance(fuentes, (list, tuple)) else fuentes
        símismo.fuentes = [_gen_fuente(f) for f in fuentes]
        símismo.variables = list(set((v for f in símismo.fuentes for v in f.variables)))

        símismo.lugares = np.unique(np.concatenate([np.unique(fnt.lugares) for fnt in símismo.fuentes]))
        símismo.fechas = np.unique(np.concatenate([np.unique(fnt.fechas) for fnt in símismo.fuentes]))

    def obt_vals(símismo, vars_interés=None, lugares=None, fechas=None):
        """
        Devuelve los valores de unos variables de interés.

        Parameters
        ----------
        vars_interés: str or list
            Los variables de interés.
        lugares: list
            Lugares de interés.
        fechas: tuple or list

        Returns
        -------
        xr.DataArray, xr.Dataset
            ``xr.DataArray`` si ``vars_interés`` es ``str``, ``xr.Dataset`` si ``vars_interés`` es ``list``.
        """
        vr_único = False
        if isinstance(vars_interés, str):
            vars_interés = [vars_interés]
            vr_único = True

        if vars_interés is None:
            vars_interés = símismo.variables

        l_vals_fnts = [
            f.obt_vals(vars_interés=vars_interés, lugares=lugares, fechas=fechas) for f in símismo.fuentes
        ]
        l_vals_fnts = [v for v in l_vals_fnts if v.data_vars]
        vals = xr.merge(l_vals_fnts)

        return vals[vars_interés[0]] if vr_único else vals

    def interpolar(símismo, vars_interés, lugares=None, fechas=None, extrap=False):
        """
        Interpola datos por fecha, tomando el lugar en cuenta.

        Parameters
        ----------
        vars_interés: str or list
            Los variables de interés.
        lugares: list
            Lugares de interés.
        fechas: list or str or datetime.datetime
            Las fechas de interés.
        extrap: bool
            Si hay que extrapolar también.

        Returns
        -------
        xr.DataArray, xr.Dataset
            ``xr.DataArray`` si ``vars_interés`` es ``str``, ``xr.Dataset`` si ``vars_interés`` es ``list``.
        """

        datos = símismo.obt_vals(vars_interés=vars_interés, lugares=lugares)
        if datos['n'].size:
            datos = _interpolar_xr(datos, fechas=fechas)

            if extrap:
                datos = datos.bfill(_('fecha'))
                datos = datos.ffill(_('fecha'))

            return datos


def _gen_fuente(fnt, nombre=None, lugares=None, fechas=None):
    if isinstance(fnt, pd.DataFrame):
        return FuentePandas(fnt, nombre or 'pandas', lugares=lugares, fechas=fechas)

    elif isinstance(fnt, dict):
        return FuenteDic(fnt, nombre or 'dic', lugares=lugares, fechas=fechas)

    elif isinstance(fnt, xr.Dataset):
        return FuenteBaseXarray(fnt, nombre or 'xarray', lugares=lugares, fechas=fechas)

    elif isinstance(fnt, xr.DataArray):
        return FuenteVarXarray(fnt, nombre or 'xarray', lugares=lugares, fechas=fechas)

    elif isinstance(fnt, str):
        ext = os.path.splitext(fnt)[1]

        if ext == '.csv':
            return FuenteCSV(fnt, lugares=lugares, fechas=fechas)
        else:
            raise ValueError(_('Formato de base de datos "{}" no reconocido.').format(ext))

    return fnt


def _interpolar_xr(m, fechas=None):
    m = m.unstack()
    if m.sizes[_('fecha')] > 1:

        m = m.interpolate_na(_('fecha')).fillna(m)
        if fechas is not None:
            fechas = [fechas] if not isinstance(fechas, (tuple, list)) else fechas
            m = m.interp(
                {_('fecha'): pd.DatetimeIndex(fechas)}
            ).fillna(m)
    m = m.dropna(_('lugar'), how='all')
    return m.stack(n=[_('lugar'), _('fecha')])



import pandas as pd
import xarray as xr
from tinamit.config import _
from تقدیر.مقام import مقام
from tinamit.tiempo.tiempo import TiempoCalendario


class Clima(object):
    def __init__(símismo, lat, long, elev=None, fuentes=None, escenario='8.5'):
        símismo._lugar = مقام(عرض=lat, طول=long, بلندی=elev)
        símismo.fuentes = fuentes
        símismo.escenario = escenario

        símismo.datos = None

    def inicializar(símismo, t):
        """
        Por hacer todas las pedidas a ``taqdir`` de una vez al principio, ahoramos mucho tiempo porque la pedida
        de datos de ``taqdir`` es bastante lenta.

        Parameters
        ----------
        t: TiempoCalendario
            El tiempo del cual querremos datos después.
        """

        f_inic = t.fecha()
        f_final = f_inic + t.delta_relativo(t.n_pasos + 1)

        if símismo.datos is None:
            símismo.datos = símismo._obt_datos_de_taqdir(f_inic, f_final)

        f_inic_datos = símismo.datos.index[0].to_timestamp()
        f_final_datos = símismo.datos.index[-1].to_timestamp()

        # En el caso que ya haya sido inicializado para otra corrida con datos distintos...
        if f_inic_datos > f_inic:
            nuevos = símismo._obt_datos_de_taqdir(f_inic, f_inic_datos - 1)
            símismo.datos = pd.concat([nuevos, símismo.datos])
        if f_final_datos < f_final:
            nuevos = símismo._obt_datos_de_taqdir(f_final_datos + 1, f_final)
            símismo.datos = pd.concat([símismo.datos, nuevos])

    def _obt_datos_de_taqdir(símismo, f_inic, f_final):
        return símismo._lugar.کوائف_پانا(
            f_inic, f_final, ذرائع=(símismo.fuentes,), خاکے=símismo.escenario
        ).روزانہ()

    def obt_datos(símismo, f_inic, f_final=None):

        f_final = f_final or f_inic
        dts = símismo.datos

        return dts[f_inic: f_final]

    def combin_datos(símismo, f_inic, f_final, vars_clima):
        vals = {}
        datos = símismo.obt_datos(f_inic, f_final)
        for v, d in vars_clima.items():
            combin = d['combin']
            datos_vr = datos[d['nombre_tqdr']]

            vals[v] = (combin(datos_vr) if combin is not None else datos_vr[-1]) * d['conv']
        return vals

    def obt_todos_vals(símismo, t, vars_clima):
        vals = {}
        datos = símismo.obt_datos(t[0], t[-1])
        for v, d in vars_clima.items():
            datos_vr = datos[d['nombre_tqdr']]
            combin = d['combin']
            conv = d['conv']
            eje = t.eje()
            datos_t = [
                combin(
                    datos_vr[(datos_vr['date'] < f) & (datos_vr['date'] <= eje[i + 1])]
                ) if combin is not None else datos_vr[f] for i, f in enumerate(eje[:-1])
            ]

            vals[v] = xr.DataArray(datos_t * conv, coords={_('fecha'): t}, dims=[_('fecha')])
        return vals



import unittest

from tinamit.config import conf, trads


class TestConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        conf['prueba'] = 1

    def test_obt_val_config(símismo):
        símismo.assertEqual(conf['prueba'], 1)

    def test_obt_val_config_anidada(símismo):
        conf['prueba2', 'prueba3'] = '¡aquí estoy!'
        símismo.assertEqual('¡aquí estoy!', conf['prueba2', 'prueba3'])

    def test_borrrar_config_no_existe(símismo):
        with símismo.assertRaises(KeyError):
            conf.borrar('Yo no existo')

    @classmethod
    def tearDownClass(cls):
        conf.borrar('prueba')
        conf.borrar('prueba2')


class TestCambiarLengua(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.leng = trads.idioma()
        cls.lengs_aux = trads.auxiliares()

    def test_cambiar_leng(símismo):
        trads.cambiar_idioma('tzj')
        símismo.assertEqual(trads.idioma(), 'tzj')

    def test_cambiar_lengs_aux(símismo):
        trads.cambiar_aux('हिं')
        símismo.assertListEqual(trads.auxiliares(), ['हिं'])

    @classmethod
    def tearDownClass(cls):
        trads.cambiar_idioma(cls.leng)
        trads.cambiar_aux(cls.lengs_aux)



from tqdm import tqdm

from tinamit.config import _
from tinamit.mod import OpsSimulGrupo
from .mod import CalibradorModSpotPy
from .valid import ValidadorMod, vars_datos_interés


class SimuladorGeog(object):
    """
    Simulador geográfico.
    """

    def __init__(símismo, mod):
        """

        Parameters
        ----------
        mod: Modelo
            El modelo para simular.
        """

        símismo.mod = mod

    def simular(símismo, t, vals_geog, vals_const=None, vars_interés=None, paralelo=False):
        """
        Efectua una simulación geográfica.

        Parameters
        ----------
        t: int or EspecTiempo
            El eje de tiempo para la simulación.
        vals_geog: dict
            Diccionario de cada lugar con sus valores de parámetros.
        vals_const: dict
            Valores de parámetros cuyos valores no cambian según el lugar.
        vars_interés: str or Variable or list
            Los variables cuyos resultados nos interesan.
        paralelo: bool
            Si se puede simular en paralelo.

        Returns
        -------
        ResultadosGrupo
        """
        vals_const = vals_const or {}

        ops = OpsSimulGrupo(
            t=t, extern=[dict(**vals_const, **vls_lg) for vls_lg in vals_geog.values()],
            vars_interés=vars_interés, nombre=list(vals_geog)
        )
        return símismo.mod.simular_grupo(ops_grupo=ops, paralelo=paralelo)


class CalibradorGeog(object):
    """
    Objeto para efectuar calibraciones geográficas.
    """

    def __init__(símismo, mod, calibrador=CalibradorModSpotPy):
        """

        Parameters
        ----------
        mod: Modelo
            El modelo para calibrar.
        calibrador: type
            Una subclase de :class:`~tinamit.calibs.mod.CalibradorMod`.
        """
        símismo.mod = mod
        símismo.calibrador = calibrador

    def calibrar(símismo, t, datos, líms_paráms, método='epm', n_iter=300, vars_obs=None, vals_geog=None,
                 vals_const=None):
        # Para hacer: acordarme de lo que quería hacer con ``vals_geog`` y ``vals_const``.
        vals_datos = datos.obt_vals()
        clbrd = símismo.calibrador(símismo.mod)

        calibs = {}
        for lg in datos.lugares:
            datos_lg = vals_datos.where(vals_datos[_('lugar')] == lg, drop=True)
            calibs[lg] = clbrd.calibrar(líms_paráms, datos=datos_lg, método=método, n_iter=n_iter, vars_obs=vars_obs)

        return calibs


class ValidadorGeog(object):
    """
    Objeto para correr validaciones de calibraciones geográficas.
    """

    def __init__(símismo, mod):
        """

        Parameters
        ----------
        mod: Modelo
            El modelo para validar.
        """
        símismo.mod = mod

    def validar(símismo, t, datos, paráms=None, funcs=None, vars_extern=None, corresp_vars=None):
        """
        Efectuar la validación.

        Parameters
        ----------
        t: int or EspecTiempo
            La especificación de tiempo para la validación.
        datos: BD
            La base de datos para la validación.
        paráms: dict
            Diccionario de los parámetros calibrados para cada lugar.
        funcs: list
            Funciones de validación para aplicar a los resultados.
        vars_extern: str or list or Variable
            Variable(s) exógenos cuyos valores se tomarán de la base de datos para alimentar la simulación y con
            los cuales por supuesto no se validará el modelo.
        corresp_vars:
            Diccionario de correspondencia entre nombres de valores en el modelo y en la base de datos.

        Returns
        -------
        dict
            Diccionario de la validación del modelo para cada variable con datos en cada lugar.
        """

        paráms = paráms or {}

        vars_interés = vars_datos_interés(símismo.mod, datos, corresp_vars=corresp_vars)

        vals_datos = datos.obt_vals(vars_interés)

        # Para hacer: utilizar `SimuladorGeog` para paralelizar y ahorar tiempo.
        valids = {}
        for lg in tqdm(datos.lugares):
            prms_lg = paráms[lg] if lg in paráms else {}
            datos_lg = vals_datos.where(vals_datos[_('lugar')] == lg, drop=True)

            if datos_lg.sizes['n']:
                valids[lg] = ValidadorMod(símismo.mod).validar(
                    t, datos=datos_lg, paráms=prms_lg, funcs=funcs, vars_extern=vars_extern, corresp_vars=corresp_vars
                )

        return valids



from .._vars import VarAuxiliar, VariablesMDS


class VariablesModVensimDLL(VariablesMDS):
    def editables(símismo):
        return [v for v in símismo if isinstance(v, VarAuxEditable)]

    def no_editables(símismo):
        return [v for v in símismo if not isinstance(v, VarAuxEditable)]


class VarAuxEditable(VarAuxiliar):
    pass



import numpy as np
import pandas as pd
import xarray as xr
from tinamit.config import _
from tinamit.datos.bd import BD
from tinamit.mod.extern import relativizar_eje
from tinamit.tiempo.tiempo import EspecTiempo

from ._utils import eval_funcs


class ValidadorMod(object):
    """
    Clase para efectuar validaciones de un modelo.
    """
    def __init__(símismo, mod):
        """

        Parameters
        ----------
        mod: Modelo
            El modelo para validar.
        """
        símismo.mod = mod

    def validar(símismo, t, datos, paráms=None, funcs=None, vars_extern=None, corresp_vars=None):
        """
        Efectua la validación.

        Parameters
        ----------
        t: int or EspecTiempo
            La especificación de tiempo para la validación.
        datos: xr.Dataset or xr.DataArray or str or pd.DataFrame or dict or Fuente or list
            La base de datos para la validación.
        paráms: dict
            Diccionario de los parámetros calibrados para cada lugar.
        funcs: list
            Funciones de validación para aplicar a los resultados.
        vars_extern: str or list or Variable
            Variable(s) exógenos cuyos valores se tomarán de la base de datos para alimentar la simulación y con
            los cuales por supuesto no se validará el modelo.
        corresp_vars:
            Diccionario de correspondencia entre nombres de valores en el modelo y en la base de datos.

        Returns
        -------
        dict
            Validación por variable.

        """

        t = t if isinstance(t, EspecTiempo) else EspecTiempo(t)
        if not isinstance(datos, xr.Dataset):
            datos = datos if isinstance(datos, BD) else BD(datos)
            datos = datos.obt_vals(
                buscar_vars_interés(símismo.mod, datos, corresp_vars)
            )

        funcs = funcs or list(eval_funcs)
        vars_extern = vars_extern or []
        if not isinstance(vars_extern, list):
            vars_extern = [vars_extern]

        vars_interés = buscar_vars_interés(símismo.mod, datos, corresp_vars)
        vars_valid = [v for v in vars_interés if v not in vars_extern]

        vals_extern = datos[list({_resolver_var(v, corresp_vars) for v in vars_extern})]
        extern = {vr: vals_extern[_resolver_var(vr, corresp_vars)].dropna('n') for vr in vars_extern}
        extern = {ll: v for ll, v in extern.items() if v.sizes['n']}

        res = símismo.mod.simular(t=t, extern={**paráms, **extern}, vars_interés=vars_valid)

        vals_calib = datos[list({_resolver_var(v, corresp_vars) for v in vars_valid})]
        # Para hacer: si implementamos Dataset en ResultadosSimul este se puede combinar en una línea
        valid = {}
        for r in res:
            vr_datos = _resolver_var(str(r), corresp_vars)
            vals_calib_vr = vals_calib[vr_datos].dropna('n')
            if vals_calib_vr.sizes['n']:
                eje = r.vals[_('fecha')].values
                eje_obs = pd.to_datetime(vals_calib_vr[_('fecha')].values)
                eje_res = relativizar_eje(eje, eje_obs)
                buenas_fechas = xr.DataArray(np.logical_and(eje_res[0] <= eje_obs, eje_obs <= eje_res[-1]), dims='n')
                datos_r = vals_calib_vr.where(buenas_fechas, drop=True).dropna('n')
                if datos_r.sizes['n'] > 1:
                    fechas_obs = datos_r[_('fecha')]
                    interpoladas = r.interpolar(fechas=fechas_obs)
                    valid[str(r)] = _valid_res(
                        datos_r.values, interpoladas.values, pd.to_datetime(datos_r[_('fecha')].values), funcs
                    )
        return valid


def _valid_res(obs, res, fechas, funcs):
    vlds = {}
    for f in funcs:
        vlds[f] = eval_funcs[f.lower()](obs, res[:, 0], fechas)  # para hacer: dimensiones múltiples
    return vlds


def _resolver_var(var, corresp_vars):
    if not corresp_vars or var not in corresp_vars:
        return var
    return corresp_vars[var]


def buscar_vars_interés(mod, datos, corresp_vars):
    return [v.nombre for v in mod.variables if _resolver_var(v.nombre, corresp_vars) in datos.variables]


def vars_datos_interés(mod, datos, corresp_vars):
    return {_resolver_var(v, corresp_vars) for v in buscar_vars_interés(mod, datos, corresp_vars=corresp_vars)}



import numpy as np

from tinamit.config import _
from tinamit.envolt.bf import VariablesModBloques, VarBloque
from tinamit.mod.var import Variable


class VarSAHYSMOD(Variable):
    def __init__(símismo, cód, **argsll):
        símismo.cód = cód
        super().__init__(**argsll)


class VarBloqSAHYSMOD(VarSAHYSMOD, VarBloque):

    def __init__(símismo, cód, nombre, unid, ingr, egr, tmñ_bloques, inic=0, líms=None, info=''):
        super().__init__(
            cód, nombre=nombre, unid=unid, ingr=ingr, egr=egr, tmñ_bloques=tmñ_bloques, inic=inic, líms=líms, info=info
        )


class VariablesSAHYSMOD(VariablesModBloques):
    def __init__(símismo, inic):
        # Guardar el número de estaciones y de polígonos
        n_estaciones = int(inic['NS'])
        dur_estaciones = [int(x) for x in inic['TS']]  # La duración de las estaciones (en meses)

        # Asegurarse que el número de estaciones es igual al número de duraciones de estaciones.
        if n_estaciones != len(dur_estaciones):
            raise ValueError(
                _('Error en el fuente de datos iniciales SAHYSMOD: el número de duraciones de estaciones'
                  'especificadas no corresponde al número de estaciones especificadas (líneas 3 y 4).')
            )

        super().__init__(_vars_sahysmod(inic, dur_est=dur_estaciones), tmñ_bloques=dur_estaciones)

    def var_a_cód(símismo, var):
        return símismo[var].cód

    def cód_a_var(símismo, cód):
        return next(v for v in símismo if v.cód.strip('#') == cód.strip('#'))


def _vars_sahysmod(d_inic, dur_est):
    n_polí = int(d_inic['NN_IN'])
    n_est = len(dur_est)
    l_vars = []
    for cód, d in _info_vars.items():
        try:
            vals = d_inic[cód.replace('#', '').upper()]
            if '#' in cód and vals.shape == (n_est, n_polí):
                inic = vals[0]
            else:
                inic = vals
                vals = None
        except KeyError:
            vals = None
            inic = np.zeros(n_polí)

        if '#' in cód:
            var = VarBloqSAHYSMOD(
                cód=cód, nombre=d['nombre'], tmñ_bloques=dur_est, unid=d['unid'], ingr=d['ingr'], egr=d['egr'],
                inic=inic
            )
            if vals is not None:
                var.poner_vals_paso(val=vals)
        else:
            var = VarSAHYSMOD(cód=cód, nombre=d['nombre'], unid=d['unid'], ingr=d['ingr'], egr=d['egr'], inic=inic)

        l_vars.append(var)

    return l_vars


# Un diccionario de variables SAHYSMOD. Ver la documentación SAHYSMOD para más detalles.
_info_vars = {
    'Pp#': {'nombre': 'Pp - Rainfall', 'unid': 'm3/season/m2', 'ingr': True, 'egr': False},
    'Ci#': {'nombre': 'Ci - Incoming canal salinity', 'unid': 'dS/m', 'ingr': True,
            'egr': True},
    'Cinf': {'nombre': 'Cinf - Aquifer inflow salinity', 'unid': 'dS/m', 'ingr': True,
             'egr': False},
    'Dc': {'nombre': 'Dc - Capillary rise critical depth', 'unid': 'm', 'ingr': True, 'egr': False},
    'Dd': {'nombre': 'Dd - Subsurface drain depth', 'unid': 'm', 'ingr': True, 'egr': False},
    'Dr': {'nombre': 'Dr - Root zone thickness', 'unid': 'm', 'ingr': True, 'egr': False},
    'Dx': {'nombre': 'Dx - Transition zone thickness', 'unid': 'm', 'ingr': True, 'egr': False},
    'EpA#': {'nombre': 'EpA - Potential ET crop A', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': False},
    'EpB#': {'nombre': 'EpB - Potential ET crop B', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': False},
    'EpU#': {'nombre': 'EpU - Potential ET non-irrigated', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': False},
    'Flq': {'nombre': 'Flq - Aquifer leaching efficienty', 'unid': 'Dmnl', 'ingr': True,
            'egr': False},
    'Flr': {'nombre': 'Flr - Root zone leaching efficiency', 'unid': 'Dmnl', 'ingr': True,
            'egr': False},
    'Flx': {'nombre': 'Flx - Transition zone leaching efficiency', 'unid': 'Dmnl', 'ingr': True,
            'egr': False},
    'Frd#': {'nombre': 'Frd - Drainage function reduction factor', 'unid': 'Dmnl', 'ingr': True,
             'egr': False},
    'FsA#': {'nombre': 'FsA - Water storage efficiency crop A', 'unid': 'Dmnl', 'ingr': True,
             'egr': False},
    'FsB#': {'nombre': 'FsB - Water storage efficiency crop B', 'unid': 'Dmnl', 'ingr': True,
             'egr': False},
    'FsU#': {'nombre': 'FsU - Water storage efficiency non-irrigated', 'unid': 'Dmnl',
             'ingr': True, 'egr': False},
    'Fw#': {'nombre': 'Fw - Fraction well water to irrigation', 'unid': 'Dmnl', 'ingr': True,
            'egr': False},
    'Gu#': {'nombre': 'Gu - Subsurface drainage for irrigation', 'unid': 'm3/season/m2',
            'ingr': True, 'egr': False},
    'Gw#': {'nombre': 'Gw - Groundwater extraction', 'unid': 'm3/season/m2', 'ingr': True,
            'egr': True},
    'Hc': {'nombre': 'Hp - Initial water level semi-confined', 'unid': 'm', 'ingr': True,
           'egr': False},
    'IaA#': {'nombre': 'IaA - Crop A field irrigation', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': True},
    'IaB#': {'nombre': 'IaB - Crop B field irrigation', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': True},
    'KcA#': {'nombre': 'Rice A - Crop A paddy?', 'unid': 'Dmnl', 'ingr': True, 'egr': False},
    'KcB#': {'nombre': 'Rice B - Crop B paddy?', 'unid': 'Dmnl', 'ingr': True, 'egr': False},
    'Kd': {'nombre': 'Kd - Subsurface drainage?', 'unid': 'Dmnl', 'ingr': True, 'egr': False},
    'Kf': {'nombre': 'Kf - Farmers\'s responses', 'unid': 'Dmnl', 'ingr': True, 'egr': False},
    'Kaq1': {'nombre': 'Kaq1 - Horizontal hydraulic conductivity 1', 'unid': 'm/day', 'ingr': True,
             'egr': False},
    'Kaq2': {'nombre': 'Kaq2 - Horizontal hydraulic conductivity 2', 'unid': 'm/day', 'ingr': True,
             'egr': False},
    'Kaq3': {'nombre': 'Kaq3 - Horizontal hydraulic conductivity 3', 'unid': 'm/day', 'ingr': True,
             'egr': False},
    'Kaq4': {'nombre': 'Kaq4 - Horizontal hydraulic conductivity 4', 'unid': 'm/day', 'ingr': True,
             'egr': False},
    'Ksc1': {'nombre': 'Ksc1 - Semi-confined aquifer 1?', 'unid': 'Dmnl', 'ingr': True,
             'egr': False},
    'Ksc2': {'nombre': 'Ksc2 - Semi-confined aquifer 2?', 'unid': 'Dmnl', 'ingr': True,
             'egr': False},
    'Ksc3': {'nombre': 'Ksc3 - Semi-confined aquifer 3?', 'unid': 'Dmnl', 'ingr': True,
             'egr': False},
    'Ksc4': {'nombre': 'Ksc4 - Semi-confined aquifer 4?', 'unid': 'Dmnl', 'ingr': True,
             'egr': False},
    'Kr#': {'nombre': 'Kr - Land use key', 'unid': 'Dmnl', 'ingr': True, 'egr': True},
    'Kvert': {'nombre': 'Kvert - Vertical hydraulic conductivity semi-confined', 'unid': 'm/day',
              'ingr': True, 'egr': False},
    'Lc#': {'nombre': 'Lc - Canal percolation', 'unid': 'm3/season/m2', 'ingr': True,
            'egr': False},
    'Peq': {'nombre': 'Peq - Aquifer effective porosity', 'unid': 'm/m', 'ingr': True,
            'egr': False},
    'Per': {'nombre': 'Per - Root zone effective porosity', 'unid': 'm/m', 'ingr': True,
            'egr': False},
    'Pex': {'nombre': 'Pex - Transition zone effective porosity', 'unid': 'm/m', 'ingr': True,
            'egr': False},
    'Psq': {'nombre': 'Psq - Semi-confined aquifer storativity', 'unid': 'Dmnl', 'ingr': True,
            'egr': False},
    'Ptq': {'nombre': 'Ptq - Aquifer total pore space', 'unid': 'm/m', 'ingr': True, 'egr': False},
    'Ptr': {'nombre': 'Ptr - Root zone total pore space', 'unid': 'm/m', 'ingr': True,
            'egr': False},
    'Ptx': {'nombre': 'Ptx - Transition zone total pore space', 'unid': 'm/m', 'ingr': True,
            'egr': False},
    'QH1': {'nombre': 'QH1 - Drain discharge to water table height ratio', 'unid': 'm/day/m',
            'ingr': True, 'egr': False},
    'QH2': {'nombre': 'QH2 - Drain discharge to sq. water table height ratio', 'unid': 'm/day/m2',
            'ingr': True, 'egr': False},
    'Qinf': {'nombre': 'Qinf - Aquifer inflow', 'unid': 'm3/season/m2', 'ingr': True, 'egr': False},
    'Qout': {'nombre': 'Qout - Aquifer outflow', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': False},
    'SL': {'nombre': 'SL - Soil surface level', 'unid': 'm', 'ingr': True, 'egr': False},
    'SiU#': {'nombre': 'SiU - Surface inflow to non-irrigated', 'unid': 'm3/season/m2',
             'ingr': True, 'egr': False},
    'SoA#': {'nombre': 'SdA - Surface outflow crop A', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': False},
    'SoB#': {'nombre': 'SdB - Surface outflow crop B', 'unid': 'm3/season/m2', 'ingr': True,
             'egr': False},
    'SoU#': {'nombre': 'SoU - Surface outflow non-irrigated', 'unid': 'm3/season/m2',
             'ingr': True, 'egr': False},
    'It#': {'nombre': 'It - Total irrigation', 'unid': 'm3/season/m2', 'ingr': False,
            'egr': True},
    'Is#': {'nombre': 'Is - Canal irrigation', 'unid': 'm3/season/m2', 'ingr': False,
            'egr': True},
    'FfA#': {'nombre': 'FfA - Irrigation efficiency crop A', 'unid': 'Dmnl', 'ingr': False,
             'egr': True},
    'FfB#': {'nombre': 'FfB - Irrigation efficiency crop B', 'unid': 'Dmnl', 'ingr': False,
             'egr': True},
    'FfT#': {'nombre': 'FfT - Total irrigation efficiency', 'unid': 'Dmnl', 'ingr': False,
             'egr': True},
    'Io#': {'nombre': 'Io - Water leaving by canal', 'unid': 'm3/season/m2', 'ingr': False,
            'egr': True},
    'JsA#': {'nombre': 'JsA - Irrigation sufficiency crop A', 'unid': 'Dmnl', 'ingr': False,
             'egr': True},
    'JsB#': {'nombre': 'JsB - Irrigation sufficiency crop B', 'unid': 'Dmnl', 'ingr': False,
             'egr': True},
    'EaU#': {'nombre': 'EaU - Actual evapotranspiration nonirrigated', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'LrA#': {'nombre': 'LrA - Root zone percolation crop A', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'LrB#': {'nombre': 'LrB - Root zone percolation crop B', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'LrU#': {'nombre': 'LrU - Root zone percolation nonirrigated', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'LrT#': {'nombre': 'LrT - Total root zone percolation', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'RrA#': {'nombre': 'RrA - Capillary rise crop A', 'unid': 'm3/season/m2', 'ingr': False,
             'egr': True},
    'RrB#': {'nombre': 'RrB - Capillary rise crop B', 'unid': 'm3/season/m2', 'ingr': False,
             'egr': True},
    'RrU#': {'nombre': 'RrU - Capillary rise non-irrigated', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'RrT#': {'nombre': 'RrT - Total capillary rise', 'unid': 'm3/season/m2', 'ingr': False,
             'egr': True},
    'Gti#': {'nombre': 'Gti - Trans zone horizontal incoming groundwater',
             'unid': 'm3/season/m2', 'ingr': False, 'egr': True},
    'Gto#': {'nombre': 'Gto - Trans zone horizontal outgoing groundwater',
             'unid': 'm3/season/m2', 'ingr': False, 'egr': True},
    'Qv#': {'nombre': 'Qv - Net vertical water table recharge', 'unid': 'm', 'ingr': False,
            'egr': True},
    'Gqi#': {'nombre': 'Gqi - Aquifer horizontal incoming groundwater', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'Gqo#': {'nombre': 'Gqo - Aquifer horizontal outgoing groundwater', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'Gaq#': {'nombre': 'Gaq - Net aquifer horizontal flow', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'Gnt#': {'nombre': 'Gnt - Net horizontal groundwater flow', 'unid': 'm3/season/m2',
             'ingr': False, 'egr': True},
    'Gd#': {'nombre': 'Gd - Total subsurface drainage', 'unid': 'm3/season/m2', 'ingr': False,
            'egr': True},
    'Ga#': {'nombre': 'Ga - Subsurface drainage above drains', 'unid': 'm3/season/m2',
            'ingr': False, 'egr': True},
    'Gb#': {'nombre': 'Gb - Subsurface drainage below drains', 'unid': 'm3/season/m2',
            'ingr': False, 'egr': True},
    'Dw#': {'nombre': 'Dw - Groundwater depth', 'unid': 'm', 'ingr': False, 'egr': True},
    'Hw#': {'nombre': 'Hw - Water table elevation', 'unid': 'm', 'ingr': True, 'egr': True},
    'Hq#': {'nombre': 'Hq - Subsoil hydraulic head', 'unid': 'm', 'ingr': False, 'egr': True},
    'Sto#': {'nombre': 'Sto - Water table storage', 'unid': 'm', 'ingr': False, 'egr': True},
    'Zs#': {'nombre': 'Zs - Surface water salt', 'unid': 'm*dS/m', 'ingr': False, 'egr': True},
    'A#': {'nombre': 'Area A - Seasonal fraction area crop A', 'unid': 'Dmnl', 'ingr': True,
           'egr': True},
    'B#': {'nombre': 'Area B - Seasonal fraction area crop B', 'unid': 'Dmnl', 'ingr': True,
           'egr': True},
    'U#': {'nombre': 'Area U - Seasonal fraction area nonirrigated', 'unid': 'Dmnl',
           'ingr': False, 'egr': True},
    'Uc#': {'nombre': 'Uc - Fraction permanently non-irrigated', 'unid': 'Dmnl', 'ingr': False,
            'egr': True},
    'CrA#': {'nombre': 'CrA - Root zone salinity crop A', 'unid': 'dS / m', 'ingr': True,
             'egr': True},
    'CrB#': {'nombre': 'CrB - Root zone salinity crop B', 'unid': 'dS / m', 'ingr': True,
             'egr': True},
    'CrU#': {'nombre': 'CrU - Root zone salinity non-irrigated', 'unid': 'dS / m', 'ingr': True,
             'egr': True},
    'Cr4#': {'nombre': 'Cr4 - Fully rotated land irrigated root zone salinity',
             'unid': 'dS / m', 'ingr': False, 'egr': True},
    'C1*#': {'nombre': 'C1 - Key 1 non-permanently irrigated root zone salinity',
             'unid': 'dS / m', 'ingr': False, 'egr': True},
    'C2*#': {'nombre': 'C2 - Key 2 non-permanently irrigated root zone salinity',
             'unid': 'dS / m', 'ingr': False, 'egr': True},
    'C3*#': {'nombre': 'C3 - Key 3 non-permanently irrigated root zone salinity',
             'unid': 'dS / m', 'ingr': False, 'egr': True},
    'Cxf#': {'nombre': 'Cxf - Transition zone salinity', 'unid': 'dS / m', 'ingr': True,
             'egr': True},
    'Cxa#': {'nombre': 'Cxa - Transition zone above-drain salinity', 'unid': 'dS / m',
             'ingr': True, 'egr': True},
    'Cxb#': {'nombre': 'Cxb - Transition zone below-drain salinity', 'unid': 'dS / m',
             'ingr': True, 'egr': True},
    'Cti#': {'nombre': 'Cti - Transition zone incoming salinity', 'unid': 'dS / m',
             'ingr': False, 'egr': True},
    'Cqf#': {'nombre': 'Cqf - Aquifer salinity', 'unid': 'dS / m', 'ingr': True, 'egr': True},
    'Cd#': {'nombre': 'Cd - Drainage salinity', 'unid': 'ds / m', 'ingr': False, 'egr': True},
    'Cw#': {'nombre': 'Cw - Well water salinity', 'unid': 'ds / m', 'ingr': False, 'egr': True}
}



import csv
import ctypes
import os
import struct

import numpy as np

from tinamit.config import _
from tinamit.cositas import arch_más_recién


def cargar_mod_vensim(mod, archivo):
    nmbr, ext = os.path.splitext(archivo)
    if ext == '.mdl':

        # Únicamente recrear el archivo .vpm si necesario
        if not os.path.isfile(nmbr + '.vpm') or arch_más_recién(archivo, nmbr + '.vpm'):
            publicar_modelo(mod=mod, archivo=archivo)
        archivo = nmbr + '.vpm'

    elif ext != '.vpm':
        raise ValueError(
            _('Vensim no sabe leer modelos del formato "{}". Debes darle un modelo ".mdl" o ".vpm".').format(ext)
        )

    # Inicializar Vensim
    cmd_vensim(func=mod.vensim_command,
               args=[''],
               mensaje_error=_('Error iniciando Vensim.'))

    # Cargar el modelo
    cmd_vensim(func=mod.vensim_command,
               args='SPECIAL>LOADMODEL|%s' % archivo,
               mensaje_error=_('Error cargando el modelo de Vensim.'))

    # Parámetros estéticos de ejecución.
    cmd_vensim(func=mod.vensim_be_quiet, args=[2],
               mensaje_error=_('Error en la comanda "vensim_be_quiet".'),
               val_error=-1)

    return mod


def inic_modelo(mod, paso, n_pasos, nombre_corrida):
    nombre_corrida = _verificar_nombre(nombre_corrida)

    # Establecer el nombre de la corrida.
    cmd_vensim(func=mod.vensim_command,
               args="SIMULATE>RUNNAME|%s" % nombre_corrida,
               mensaje_error=_('Error iniciando la corrida Vensim.'))

    # Establecer el tiempo final.
    cmd_vensim(func=mod.vensim_command,
               args='SIMULATE>SETVAL|%s = %i' % ('FINAL TIME', n_pasos + 1),
               mensaje_error=_('Error estableciendo el tiempo final para Vensim.'))

    # Iniciar la simulación en modo juego ("Game"). Esto permitirá la actualización de los valores de los variables
    # a través de la simulación.
    cmd_vensim(func=mod.vensim_command,
               args="MENU>GAME",
               mensaje_error=_('Error inicializando el juego Vensim.'))

    # Es ABSOLUTAMENTE necesario establecer el intervalo del juego aquí. Sino, no reinicializa el paso
    # correctamente entre varias corridas (aún modelos) distintas.
    cmd_vensim(func=mod.vensim_command,
               args="GAME>GAMEINTERVAL|%i" % paso,
               mensaje_error=_('Error estableciendo el paso de Vensim.'))


def avanzar_modelo(mod):
    # Avanzar el modelo.
    cmd_vensim(func=mod.vensim_command,
               args="GAME>GAMEON", mensaje_error=_('Error en incrementar Vensim.'))


def estab_paso(mod, paso):
    cmd_vensim(func=mod.vensim_command,
               args="GAME>GAMEINTERVAL|%i" % paso,
               mensaje_error=_('Error estableciendo el paso de Vensim.'))


def obt_paso_inicial(mod):
    # para hacer: verificar
    try:
        return obt_val_var(mod, 'Game Interval', subs=None)
    except ValueError:
        return obt_val_var(mod, 'TIME STEP', subs=None)


def cambiar_val(mod, var, val):
    cmd_vensim(func=mod.vensim_command,
               args='SIMULATE>SETVAL|%s = %f' % (var, val),
               mensaje_error=_('Error cambiando el variable %s.') % var)


def verificar_vensim(símismo):
    """
    Esta función regresa el estatus de Vensim. Es particularmente útil para desboguear (no tiene uso en las
    otras funciones de esta clase, y se incluye como ayuda a la programadora.)

    :return: Código de estatus Vensim:
        | 0 = Vensim está listo
        | 1 = Vensim está en una simulación activa
        | 2 = Vensim está en una simulación, pero no está respondiendo
        | 3 = Malas noticias
        | 4 = Error de memoria
        | 5 = Vensim está en modo de juego
        | 6 = Memoria no libre. Llamar vensim_command() debería de arreglarlo.
        | 16 += ver documentación de Vensim para vensim_check_status() en la sección de DLL (Suplemento DSS)
    :rtype: int

    """

    # Obtener el estatus.
    estatus = cmd_vensim(func=símismo.mod.vensim_check_status,
                         args=[],
                         mensaje_error=_('Error verificando el estatus de Vensim. De verdad, la cosa '
                                         'te va muy mal.'),
                         val_error=-1)
    return int(estatus)


def publicar_modelo(mod, archivo):  # pragma: sin cobertura

    cmd_vensim(mod.vensim_command, 'SPECIAL>LOADMODEL|%s' % archivo)

    archivo_frm = os.path.join(os.path.split(os.path.dirname(__file__))[0], 'VENSIM.frm')
    cmd_vensim(mod.vensim_command, ('FILE>PUBLISH|%s' % archivo_frm))


def obt_vars(mod):
    l_nombres = []
    for t in [1, 2, 4, 5, 12]:
        mem = ctypes.create_string_buffer(0)  # Crear una memoria intermedia

        # Verificar el tamaño necesario
        tamaño_nec = cmd_vensim(
            func=mod.vensim_get_varnames,
            args=['*', t, mem, 0],
            mensaje_error=_('Error obteniendo eñ tamaño de los variables Vensim.'),
            val_error=-1
        )

        mem = ctypes.create_string_buffer(tamaño_nec)  # Una memoria intermedia con el tamaño apropiado

        # Guardar y decodar los nombres de los variables.
        cmd_vensim(func=mod.vensim_get_varnames,
                   args=['*', t, mem, tamaño_nec],
                   mensaje_error=_('Error obteniendo los nombres de los variables de Vensim.'),
                   val_error=-1
                   )
        l_nombres += [
            x for x in mem.raw.decode().split('\x00')
            if x and x not in ['FINAL TIME', 'TIME STEP', 'INITIAL TIME', 'SAVEPER', 'Time']
        ]
    return l_nombres


def obt_editables(mod):
    # Para obtener los nombres de variables editables (se debe hacer así y no por `tipo_var` porque Vensim
    # los reporta como de tipo `Auxiliary`.
    mem = ctypes.create_string_buffer(0)  # Crear una memoria intermedia

    # Verificar el tamaño necesario
    tamaño_nec = cmd_vensim(
        func=mod.vensim_get_varnames,
        args=['*', 12, mem, 0],
        mensaje_error=_('Error obteniendo eñ tamaño de los variables Vensim.'),
        val_error=-1
    )

    mem = ctypes.create_string_buffer(tamaño_nec)  # Una memoria intermedia con el tamaño apropiado

    cmd_vensim(
        func=mod.vensim_get_varnames,
        args=['*', 12, mem, tamaño_nec],
        mensaje_error=_('Error obteniendo los nombres de los variables editables ("Gaming") de '
                        'VENSIM.'),
        val_error=-1
    )

    return [x for x in mem.raw.decode().split('\x00') if x]


def obt_unid_tiempo(mod):
    return obt_atrib_var(
        mod, var='TIME STEP', cód_attrib=1,
        mns_error=_('Error obteniendo la unidad de tiempo para el modelo Vensim.')
    )


def _obt_val_var_unidim(mod, var):
    # Una memoria
    mem_inter = ctypes.create_string_buffer(4)

    # Leer el valor del variable
    cmd_vensim(func=mod.vensim_get_val,
               args=[str(var), mem_inter],
               mensaje_error=_('Error con Vensim para leer variable "{}".').format(var))

    # Decodar
    return struct.unpack('f', mem_inter)[0]


def vdf_a_csv(mod, archivo_csv):

    # Vensim hace la conversión para nosotr@s
    cmd_vensim(
        mod.vensim_command,
        'MENU>VDF2CSV|{archVDF}|{archCSV}'.format(
            archVDF='!.vdf', archCSV='!.csv'  # En Vensim, "!" quiere decir la corrida activa
        ).encode()
    )

    # Cortar el último paso de simulación. Tinamït siempre corre simulaciones de Vensim para 1 paso adicional
    # para permitir que valores de variables conectados se puedan actualizar.
    # Para que quede claro: esto es por culpa de un error en Vensim, no es culpa nuestra.
    with open(archivo_csv + '.csv', 'r', encoding='UTF-8') as d:
        filas = [f[:-1] if len(f) > 2 else f for f in csv.reader(d)]

    # Hay que abrir el archivo de nuevo para re-escribir sobre el contenido existente-
    with open(archivo_csv + '.csv', 'w', encoding='UTF-8', newline='') as d:
        csv.writer(d).writerows(filas)


def cerrar_vensim(mod):
    # Necesario para guardar los últimos valores de los variables conectados. (Muy incómodo, yo sé.)
    estab_paso(mod, 1)  # justo acaso

    cmd_vensim(func=mod.vensim_command,
               args="GAME>GAMEON",
               mensaje_error=_('Error terminando la simulación Vensim.'))

    # ¡Por fin! Llamar la comanda para terminar la simulación.
    cmd_vensim(func=mod.vensim_command,
               args="GAME>ENDGAME",
               mensaje_error=_('Error terminando la simulación Vensim.'))


def obt_atrib_var(mod, var, cód_attrib, mns_error=None):
    if cód_attrib in [4, 5, 6, 7, 8, 9, 10]:
        lista = True
    elif cód_attrib in [1, 2, 3, 11, 12, 13, 14]:
        lista = False
    else:
        raise ValueError(_('Código "{}" no reconocido para la comanda Vensim de obtener atributos de variables.')
                         .format(cód_attrib))

    if mns_error is None:
        l_atrs = [_('las unidades'), _('la descipción'), _('la ecuación'), _('las causas'), _('las consecuencias'),
                  _('la causas iniciales'), _('las causas activas'), _('los subscriptos'),
                  _('las combinaciones de subscriptos'), _('los subscriptos de gráfico'), _('el mínimo'),
                  _('el máximo'), _('el rango'), _('el tipo_mod')]
        mns_error1 = _('Error leyendo el tamaño de memoria para obtener {} del variable "{}" en Vensim') \
            .format(l_atrs[cód_attrib - 1], var)
        mns_error2 = _('Error leyendo {} del variable "{}" en Vensim.').format(l_atrs[cód_attrib - 1], var)
    else:
        mns_error1 = mns_error2 = mns_error

    mem = ctypes.create_string_buffer(10)
    tmñ = cmd_vensim(
        func=mod.vensim_get_varattrib,
        args=[var, cód_attrib, mem, 0],
        mensaje_error=mns_error1,
        val_error=-1,
    )

    mem = ctypes.create_string_buffer(tmñ)
    cmd_vensim(func=mod.vensim_get_varattrib,
               args=[var, cód_attrib, mem, tmñ],
               mensaje_error=mns_error2,
               val_error=-1)

    if lista:
        return [x for x in mem.raw.decode().split('\x00') if x]
    else:
        return mem.value.decode()


def cmd_vensim(func, args, mensaje_error=None, val_error=None):  # pragma: sin cobertura
    """
    Esta función sirve para llamar todo tipo_mod de comanda Vensim.

    Parameters
    ----------
    func: callable
        La función DLL a llamar.
    args: list | str
        Los argumento a pasar a la función. Si no hay, usar una lista vacía.
    mensaje_error: str
        El mensaje de error para mostrar si hay un error en la comanda.
    val_error: int
        Un valor de regreso Vensim que indica un error para esta función. Si se deja ``None``, todos
        valores que no son 1 se considerarán como erróneas.

    Returns
    -------
    int
        El código devuelto por Vensim.
    """

    # Asegurarse que args es una lista
    if type(args) is not list:
        args = [args]

    # Encodar en bytes todos los argumentos de texto.
    for n, a in enumerate(args):
        if type(a) is str:
            args[n] = a.encode()

    # Llamar la función Vensim y guardar el resultado.
    try:
        resultado = func(*args)
    except OSError:
        try:
            resultado = func(*args)
        except OSError as e:
            raise OSError(e)

    # Verificar su hubo un error.
    if val_error is None:
        error = (resultado != 1)
    else:
        error = (resultado == val_error)

    # Si hubo un error, avisar el usuario.
    if error:
        if mensaje_error is None:
            mensaje_error = _('Error con la comanda Vensim.')

        mensaje_error += _(' Código de error {}.').format(resultado)

        raise ValueError(mensaje_error)

    # Devolver el valor devuelto por la función Vensim
    return resultado


def _verificar_nombre(nombre):
    # Vensim tiene un problema raro con nombres de corridas con '.' (y quién sabe qué más)
    return nombre.replace('.', '_')


def obt_val_var(mod, var, subs):
    if not subs:
        # Si el variable no tiene dimensiones (subscriptos)...

        # Leer su valor.
        val = _obt_val_var_unidim(mod, var)

    else:
        val = np.zeros_like(subs)
        for n, s in enumerate(subs):
            var_s = str(var) + s

            # Leer su valor.
            v = _obt_val_var_unidim(mod, var_s)

            # Guardar en el diccionario interno.
            val[n] = v  # Para hacer: opciones de dimensiones múltiples

    return val
